# Copyright 2013: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Rally command: task"""

from __future__ import print_function
import collections
import datetime as dt
import itertools
import json
import os
import sys
import webbrowser

import jsonschema

from rally.cli import cliutils
from rally.cli import envutils
from rally.common import fileutils
from rally.common import logging
from rally.common import utils as rutils
from rally.common import version
from rally.common import yamlutils as yaml
from rally import consts
from rally import exceptions
from rally import plugins
from rally.task import atomic
from rally.task.processing import charts
from rally.task import utils as tutils
from rally.utils import strutils


LOG = logging.getLogger(__name__)

OLD_TASK_RESULT_SCHEMA = {
    "type": "object",
    "$schema": consts.JSON_SCHEMA,
    "properties": {
        "key": {
            "type": "object",
            "properties": {
                "kw": {
                    "type": "object"
                },
                "name": {
                    "type": "string"
                },
                "pos": {
                    "type": "integer"
                },
            },
            "required": ["kw", "name", "pos"]
        },
        "sla": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "criterion": {
                        "type": "string"
                    },
                    "detail": {
                        "type": "string"
                    },
                    "success": {
                        "type": "boolean"
                    }
                }
            }
        },
        "hooks": {"type": "array"},
        "result": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "atomic_actions": {
                        "type": "object"
                    },
                    "duration": {
                        "type": "number"
                    },
                    "error": {
                        "type": "array"
                    },
                    "idle_duration": {
                        "type": "number"
                    },
                    "output": {"type": "object"}
                },
                "required": ["atomic_actions", "duration", "error",
                             "idle_duration"]
            },
            "minItems": 0
        },
        "load_duration": {
            "type": "number",
        },
        "full_duration": {
            "type": "number",
        },
        "created_at": {
            "type": "string"
        }
    },
    "required": ["key", "sla", "result", "load_duration", "full_duration"],
    "additionalProperties": False
}


class FailedToLoadTask(exceptions.RallyException):
    error_code = 117
    msg_fmt = "Invalid %(source)s passed:\n\n\t %(msg)s"


class FailedToLoadResults(exceptions.RallyException):
    error_code = 225
    msg_fmt = "ERROR: Invalid task result format in %(source)s\n\n\t%(msg)s"


class TaskCommands(object):
    """Set of commands that allow you to manage tasks and results.

    """

    def _load_and_validate_task(self, api, task_file, args_file=None,
                                raw_args=None):
        """Load, render and validate tasks template from file with passed args.

        :param task_file: Path to file with input task
        :param raw_args: JSON or YAML representation of dict with args that
            will be used to render input task with jinja2
        :param args_file: Path to file with JSON or YAML representation
            of dict, that will be used to render input with jinja2. If both
            specified task_args and task_args_file they will be merged.
            raw_args has bigger priority so it will update values
            from args_file.
        :returns: Str with loaded and rendered task
        """

        print(cliutils.make_header("Preparing input task"))

        try:
            input_task = open(task_file).read()
        except IOError as err:
            raise FailedToLoadTask(
                source="--task",
                msg="Error reading %s: %s" % (task_file, err))

        task_dir = os.path.expanduser(os.path.dirname(task_file)) or "./"

        task_args = {}
        if args_file:
            try:
                task_args.update(yaml.safe_load(open(args_file).read()))
            except yaml.ParserError as e:
                raise FailedToLoadTask(
                    source="--task-args-file",
                    msg="File '%s' has to be YAML or JSON. Details:\n\n%s"
                    % (args_file, e))
            except IOError as err:
                raise FailedToLoadTask(
                    source="--task-args-file",
                    msg="Error reading %s: %s" % (args_file, err))

        if raw_args:
            try:
                data = yaml.safe_load(raw_args)
                if isinstance(data, str):
                    raise yaml.ParserError("String '%s' doesn't look like a "
                                           "dictionary." % raw_args)
                task_args.update(data)
            except yaml.ParserError as e:
                args = [keypair.split("=", 1)
                        for keypair in raw_args.split(",")]
                if len([a for a in args if len(a) != 1]) != len(args):
                    raise FailedToLoadTask(
                        source="--task-args",
                        msg="Value has to be YAML or JSON. Details:\n\n%s" % e)
                else:
                    task_args.update(dict(args))

        try:
            rendered_task = api.task.render_template(task_template=input_task,
                                                     template_dir=task_dir,
                                                     **task_args)
        except Exception as e:
            raise FailedToLoadTask(
                source="--task",
                msg="Failed to render task template.\n\n%s" % e)

        print("Task is:\n%s\n" % rendered_task.strip())
        try:
            parsed_task = yaml.safe_load(rendered_task)
        except Exception as e:
            raise FailedToLoadTask(
                source="--task",
                msg="Wrong format of rendered input task. It should be YAML or"
                    " JSON. Details:\n\n%s" % e)

        print("Task syntax is correct :)")
        return parsed_task

    @cliutils.args("--deployment", dest="deployment", type=str,
                   metavar="<uuid>", required=False,
                   help="UUID or name of a deployment.")
    @cliutils.args("--task", "--filename", metavar="<path>",
                   dest="task_file",
                   help="Path to the input task file.")
    @cliutils.args("--task-args", metavar="<json>", dest="task_args",
                   help="Input task args (JSON dict). These args are used "
                        "to render the Jinja2 template in the input task.")
    @cliutils.args("--task-args-file", metavar="<path>", dest="task_args_file",
                   help="Path to the file with input task args (dict in "
                        "JSON/YAML). These args are used "
                        "to render the Jinja2 template in the input task.")
    @envutils.with_default_deployment(cli_arg_name="deployment")
    @plugins.ensure_plugins_are_loaded
    def validate(self, api, task_file, deployment=None, task_args=None,
                 task_args_file=None):
        """Validate a task configuration file.

        This will check that task configuration file has valid syntax and
        all required options of scenarios, contexts, SLA and runners are set.

        If both task_args and task_args_file are specified, they will
        be merged. task_args has a higher priority so it will override
        values from task_args_file.
        """

        task = self._load_and_validate_task(api, task_file, raw_args=task_args,
                                            args_file=task_args_file)

        api.task.validate(deployment=deployment, config=task)

        print("Input Task is valid :)")

    @cliutils.args("--deployment", dest="deployment", type=str,
                   metavar="<uuid>", required=False,
                   help="UUID or name of a deployment.")
    @cliutils.args("--task", "--filename", metavar="<path>",
                   dest="task_file",
                   help="Path to the input task file.")
    @cliutils.args("--task-args", dest="task_args", metavar="<json>",
                   help="Input task args (JSON dict). These args are used "
                        "to render the Jinja2 template in the input task.")
    @cliutils.args("--task-args-file", dest="task_args_file", metavar="<path>",
                   help="Path to the file with input task args (dict in "
                        "JSON/YAML). These args are used "
                        "to render the Jinja2 template in the input task.")
    @cliutils.args("--tag", nargs="+", dest="tags", type=str, required=False,
                   help="Mark the task with a tag or a few tags.")
    @cliutils.args("--no-use", action="store_false", dest="do_use",
                   help="Don't set new task as default for future operations.")
    @cliutils.args("--abort-on-sla-failure", action="store_true",
                   dest="abort_on_sla_failure",
                   help="Abort the execution of a task when any SLA check "
                        "for it fails for subtask or workload.")
    @envutils.with_default_deployment(cli_arg_name="deployment")
    @plugins.ensure_plugins_are_loaded
    def start(self, api, task_file, deployment=None, task_args=None,
              task_args_file=None, tags=None, do_use=False,
              abort_on_sla_failure=False):
        """Run task.

        If both task_args and task_args_file are specified, they are going to
        be merged. task_args has a higher priority so it overrides
        values from task_args_file.
        There are 3 kinds of return codes, 0: no error, 1: running error,
        2: sla check failed.
        """

        input_task = self._load_and_validate_task(api, task_file,
                                                  raw_args=task_args,
                                                  args_file=task_args_file)
        print("Running Rally version", version.version_string())

        try:
            task_instance = api.task.create(deployment=deployment, tags=tags)
            tags = "[tags: '%s']" % "', '".join(tags) if tags else ""

            print(cliutils.make_header(
                "Task %(tags)s %(uuid)s: started"
                % {"uuid": task_instance["uuid"], "tags": tags}))
            print("Running Task... This can take a while...\n")
            print("To track task status use:\n")
            print("\trally task status\n\tor\n\trally task detailed\n")

            if do_use:
                self.use(api, task_instance["uuid"])

            api.task.start(deployment=deployment, config=input_task,
                           task=task_instance["uuid"],
                           abort_on_sla_failure=abort_on_sla_failure)

        except exceptions.DeploymentNotFinishedStatus as e:
            print("Cannot start a task on unfinished deployment: %s" % e)
            return 1

        if self._detailed(api, task_id=task_instance["uuid"]):
            return 2
        return 0

    @cliutils.args("--uuid", type=str, dest="task_id", help="UUID of task.")
    @envutils.with_default_task_id
    @cliutils.args(
        "--soft", action="store_true",
        help="Abort task after current scenario finishes execution.")
    def abort(self, api, task_id=None, soft=False):
        """Abort a running task."""

        if soft:
            print("INFO: please be informed that soft abort won't stop "
                  "a running workload, but will prevent new ones from "
                  "starting. If you are running task with only one "
                  "scenario, soft abort will not help at all.")

        api.task.abort(task_uuid=task_id, soft=soft, wait=True)

        print("Task %s successfully stopped." % task_id)

    @cliutils.args("--uuid", type=str, dest="task_id", help="UUID of task")
    @envutils.with_default_task_id
    def status(self, api, task_id=None):
        """Display the current status of a task."""

        task = api.task.get(task_id=task_id)
        print("Task %(task_id)s: %(status)s"
              % {"task_id": task_id, "status": task["status"]})

    @cliutils.args("--uuid", type=str, dest="task_id",
                   help=("UUID of task. If --uuid is \"last\" the results of "
                         " the most recently created task will be displayed."))
    @cliutils.args("--iterations-data", dest="iterations_data",
                   action="store_true",
                   help="Print detailed results for each iteration.")
    @cliutils.args("--filter-by", dest="filters", nargs="+", type=str,
                   help="Filter the displayed workloads."
                        "<sla-failures>: only display the failed workloads.\n"
                        "<scenarios>: filter the workloads by scenarios.,"
                        "scenarios=scenario_name1[,scenario_name2]...")
    @envutils.with_default_task_id
    def detailed(self, api, task_id=None, iterations_data=False,
                 filters=None):
        self._detailed(api, task_id, iterations_data, filters)

    def _detailed(self, api, task_id=None, iterations_data=False,
                  filters=None):
        """Print detailed information about given task."""
        scenarios_filter = []
        only_sla_failures = False
        for filter in filters or []:
            if filter.startswith("scenario="):
                filter_value = filter.split("=")[1]
                scenarios_filter = filter_value.split(",")
            if filter == "sla-failures":
                only_sla_failures = True

        task = api.task.get(task_id=task_id, detailed=True)

        print()
        print("-" * 80)
        print("Task %(task_id)s: %(status)s"
              % {"task_id": task_id, "status": task["status"]})

        if task["status"] == consts.TaskStatus.CRASHED or task["status"] == (
                consts.TaskStatus.VALIDATION_FAILED):
            print("-" * 80)
            validation = task["validation_result"]
            if logging.is_debug():
                print(yaml.safe_load(validation["trace"]))
            else:
                print(validation["etype"])
                print(validation["msg"])
                print("\nFor more details run:\nrally -d task detailed %s"
                      % task["uuid"])
            return 0
        elif task["status"] not in [consts.TaskStatus.FINISHED,
                                    consts.TaskStatus.ABORTED]:
            print("-" * 80)
            print("\nThe task %s marked as '%s'. Results "
                  "available when it is '%s'."
                  % (task_id, task["status"], consts.TaskStatus.FINISHED))
            return 0

        for workload in itertools.chain(
                *[s["workloads"] for s in task["subtasks"]]):
            if scenarios_filter and workload["name"] not in scenarios_filter:
                continue
            if only_sla_failures and workload["pass_sla"]:
                continue

            print("-" * 80)
            print()
            print("test scenario %s" % workload["name"])
            print("args position %s" % workload["position"])
            print("args values:")
            print(json.dumps(
                {"args": workload["args"],
                 "runner": workload["runner"],
                 "contexts": workload["contexts"],
                 "sla": workload["sla"],
                 "hooks": [r["config"] for r in workload["hooks"]]},
                indent=2))
            print()

            duration_stats = workload["statistics"]["durations"]

            iterations = []
            iterations_headers = ["iteration", "duration"]
            iterations_actions = []
            output = []
            task_errors = []
            if iterations_data:
                atomic_names = [a["display_name"]
                                for a in duration_stats["atomics"]]
                for i, atomic_name in enumerate(atomic_names, 1):
                    action = "%i. %s" % (i, atomic_name)
                    iterations_headers.append(action)
                    iterations_actions.append((atomic_name, action))

            for idx, itr in enumerate(workload["data"], 1):

                if iterations_data:
                    row = {"iteration": idx, "duration": itr["duration"]}
                    for name, action in iterations_actions:
                        atomic_actions = atomic.merge_atomic_actions(
                            itr["atomic_actions"])
                        row[action] = atomic_actions.get(name, {}).get(
                            "duration", 0)
                    iterations.append(row)

                if "output" in itr:
                    iteration_output = itr["output"]
                else:
                    iteration_output = {"additive": [], "complete": []}

                for idx, additive in enumerate(iteration_output["additive"]):
                    if len(output) <= idx + 1:
                        output_table = charts.OutputStatsTable(
                            workload, title=additive["title"])
                        output.append(output_table)
                    output[idx].add_iteration(additive["data"])

                if itr.get("error"):
                    task_errors.append(TaskCommands._format_task_error(itr))

            self._print_task_errors(task_id, task_errors)

            cols = charts.MainStatsTable.columns
            formatters = {
                "Action": lambda x: x["display_name"],
                "Min (sec)": lambda x: x["data"]["min"],
                "Median (sec)": lambda x: x["data"]["median"],
                "90%ile (sec)": lambda x: x["data"]["90%ile"],
                "95%ile (sec)": lambda x: x["data"]["95%ile"],
                "Max (sec)": lambda x: x["data"]["max"],
                "Avg (sec)": lambda x: x["data"]["avg"],
                "Success": lambda x: x["data"]["success"],
                "Count": lambda x: x["data"]["iteration_count"]
            }

            rows = []

            def make_flat(r, depth=0):
                if depth > 0:
                    r["display_name"] = (" %s> %s" % ("-" * depth,
                                                      r["display_name"]))

                rows.append(r)
                for children in r["children"]:
                    make_flat(children, depth + 1)

            for row in itertools.chain(duration_stats["atomics"],
                                       [duration_stats["total"]]):
                make_flat(row)
            cliutils.print_list(rows,
                                fields=cols,
                                formatters=formatters,
                                normalize_field_names=True,
                                table_label="Response Times (sec)",
                                sortby_index=None)
            print()

            if iterations_data:
                formatters = dict(zip(iterations_headers[1:],
                                      [cliutils.pretty_float_formatter(col, 3)
                                       for col in iterations_headers[1:]]))
                cliutils.print_list(iterations,
                                    fields=iterations_headers,
                                    table_label="Atomics per iteration",
                                    formatters=formatters)
                print()

            if output:
                cols = charts.OutputStatsTable.columns
                float_cols = cols[1:7]
                formatters = dict(zip(float_cols,
                                  [cliutils.pretty_float_formatter(col, 3)
                                   for col in float_cols]))

                for out in output:
                    data = out.render()
                    rows = [dict(zip(cols, r)) for r in data["data"]["rows"]]
                    if rows:
                        # NOTE(amaretskiy): print title explicitly because
                        #     prettytable fails if title length is too long
                        print(data["title"])
                        cliutils.print_list(rows, fields=cols,
                                            formatters=formatters)
                        print()

            print("Load duration: %s"
                  % strutils.format_float_to_str(workload["load_duration"]))
            print("Full duration: %s"
                  % strutils.format_float_to_str(workload["full_duration"]))

        print("\nHINTS:")
        print("* To plot HTML graphics with this data, run:")
        print("\trally task report %s --out output.html\n" % task["uuid"])
        print("* To generate a JUnit report, run:")
        print("\trally task export %s --type junit-xml --to output.xml\n" %
              task["uuid"])
        print("* To get raw JSON output of task results, run:")
        print("\trally task report %s --json --out output.json\n" %
              task["uuid"])

        if not task["pass_sla"]:
            print("At least one workload did not pass SLA criteria.\n")
            return 1

    @cliutils.args("--uuid", type=str, dest="task_id", help="UUID of task.")
    @envutils.with_default_task_id
    @cliutils.suppress_warnings
    def results(self, api, task_id=None):
        """Display raw task results.

        This will produce a lot of output data about every iteration.
        """

        task = api.task.get(task_id=task_id, detailed=True)
        finished_statuses = (consts.TaskStatus.FINISHED,
                             consts.TaskStatus.ABORTED)
        if task["status"] not in finished_statuses:
            print("Task status is %s. Results available when it is one of %s."
                  % (task["status"], ", ".join(finished_statuses)))
            return 1

        # TODO(chenhb): Ensure `rally task results` puts out old format.
        for workload in itertools.chain(
                *[s["workloads"] for s in task["subtasks"]]):
            for itr in workload["data"]:
                itr["atomic_actions"] = collections.OrderedDict(
                    tutils.WrapperForAtomicActions(
                        itr["atomic_actions"]).items()
                )

        results = []
        for w in itertools.chain(*[s["workloads"] for s in task["subtasks"]]):
            w["runner"]["type"] = w["runner_type"]

            def port_hook_cfg(h):
                h["config"] = {
                    "name": h["config"]["action"][0],
                    "args": h["config"]["action"][1],
                    "description": h["config"].get("description", ""),
                    "trigger": {"name": h["config"]["trigger"][0],
                                "args": h["config"]["trigger"][1]}
                }
                return h

            hooks = [port_hook_cfg(h) for h in w["hooks"]]

            created_at = dt.datetime.strptime(w["created_at"],
                                              "%Y-%m-%dT%H:%M:%S")
            created_at = created_at.strftime("%Y-%d-%mT%H:%M:%S")

            results.append({
                "key": {
                    "name": w["name"],
                    "description": w["description"],
                    "pos": w["position"],
                    "kw": {
                        "args": w["args"],
                        "runner": w["runner"],
                        "context": w["contexts"],
                        "sla": w["sla"],
                        "hooks": [h["config"] for h in w["hooks"]],
                    }
                },
                "result": w["data"],
                "sla": w["sla_results"].get("sla", []),
                "hooks": hooks,
                "load_duration": w["load_duration"],
                "full_duration": w["full_duration"],
                "created_at": created_at})

        print(json.dumps(results, sort_keys=False, indent=4))

    @cliutils.args("--deployment", dest="deployment", type=str,
                   metavar="<uuid>", required=False,
                   help="UUID or name of a deployment.")
    @cliutils.args("--all-deployments", action="store_true",
                   dest="all_deployments",
                   help="List tasks from all deployments.")
    @cliutils.args("--status", type=str, dest="status",
                   help="List tasks with specified status."
                   " Available statuses: %s" % ", ".join(consts.TaskStatus))
    @cliutils.args("--tag", nargs="+", dest="tags", type=str, required=False,
                   help="Tags to filter tasks by.")
    @cliutils.args("--uuids-only", action="store_true",
                   dest="uuids_only", help="List task UUIDs only.")
    @envutils.with_default_deployment(cli_arg_name="deployment")
    def list(self, api, deployment=None, all_deployments=False, status=None,
             tags=None, uuids_only=False):
        """List tasks, started and finished.

        Displayed tasks can be filtered by status or deployment.  By
        default 'rally task list' will display tasks from the active
        deployment without filtering by status.
        """

        filters = {}
        headers = ["UUID", "Deployment name", "Created at", "Load duration",
                   "Status", "Tag(s)"]

        if status in consts.TaskStatus:
            filters["status"] = status
        elif status:
            print("Error: Invalid task status '%s'.\nAvailable statuses: %s"
                  % (status, ", ".join(consts.TaskStatus)),
                  file=sys.stderr)
            return(1)

        if not all_deployments:
            filters["deployment"] = deployment

        if tags:
            filters["tags"] = tags

        task_list = api.task.list(**filters)

        if uuids_only:
            if task_list:
                print("\n".join([t["uuid"] for t in task_list]))
        elif task_list:
            def tags_formatter(t):
                if not t["tags"]:
                    return ""
                return "'%s'" % "', '".join(t["tags"])

            formatters = {
                "Tag(s)": tags_formatter,
                "Load duration": cliutils.pretty_float_formatter(
                    "task_duration", 3),
                "Created at": lambda t: t["created_at"].replace("T", " ")
            }

            cliutils.print_list(
                task_list, fields=headers, normalize_field_names=True,
                sortby_index=headers.index("Created at"),
                formatters=formatters)
        else:
            if status:
                print("There are no tasks in '%s' status. "
                      "To run a new task, use:\n\trally task start"
                      % status)
            else:
                print("There are no tasks. To run a new task, use:\n"
                      "\trally task start")

    def _load_task_results_file(self, api, task_id):
        """Load the json file which is created by `rally task results`"""

        with open(os.path.expanduser(task_id)) as inp_js:
            tasks_results = json.loads(inp_js.read())

        if isinstance(tasks_results, list):
            # it is an old format:

            task = {"version": 2,
                    "title": "Task loaded from a file.",
                    "description": "Auto-ported from task format V1.",
                    "uuid": "n/a",
                    "env_name": "n/a",
                    "env_uuid": "n/a",
                    "tags": [],
                    "status": consts.TaskStatus.FINISHED,
                    "subtasks": []}

            start_time = None

            for result in tasks_results:
                try:
                    jsonschema.validate(
                        result, OLD_TASK_RESULT_SCHEMA)
                except jsonschema.ValidationError as e:
                    raise FailedToLoadResults(source=task_id,
                                              msg=str(e))

                iter_count = 0
                failed_iter_count = 0
                min_duration = None
                max_duration = None

                for itr in result["result"]:
                    if start_time is None or itr["timestamp"] < start_time:
                        start_time = itr["timestamp"]
                    # NOTE(chenhb): back compatible for atomic_actions
                    itr["atomic_actions"] = list(
                        tutils.WrapperForAtomicActions(itr["atomic_actions"],
                                                       itr["timestamp"]))

                    iter_count += 1
                    if itr.get("error"):
                        failed_iter_count += 1

                    duration = itr.get("duration", 0)

                    if max_duration is None or duration > max_duration:
                        max_duration = duration

                    if min_duration is None or min_duration > duration:
                        min_duration = duration

                durations_stat = charts.MainStatsTable(
                    {"total_iteration_count": iter_count})

                for itr in result["result"]:
                    durations_stat.add_iteration(itr)

                created_at = dt.datetime.strptime(result["created_at"],
                                                  "%Y-%d-%mT%H:%M:%S")
                updated_at = created_at + dt.timedelta(
                    seconds=result["full_duration"])
                created_at = created_at.strftime(consts.TimeFormat.ISO8601)
                updated_at = updated_at.strftime(consts.TimeFormat.ISO8601)
                pass_sla = all(s.get("success") for s in result["sla"])
                runner_type = result["key"]["kw"]["runner"].pop("type")
                for h in result["hooks"]:
                    trigger = h["config"]["trigger"]
                    h["config"] = {
                        "description": h["config"].get("description"),
                        "action": (h["config"]["name"], h["config"]["args"]),
                        "trigger": (trigger["name"], trigger["args"])}
                workload = {"uuid": "n/a",
                            "name": result["key"]["name"],
                            "position": result["key"]["pos"],
                            "description": result["key"].get("description",
                                                             ""),
                            "full_duration": result["full_duration"],
                            "load_duration": result["load_duration"],
                            "total_iteration_count": iter_count,
                            "failed_iteration_count": failed_iter_count,
                            "min_duration": min_duration,
                            "max_duration": max_duration,
                            "start_time": start_time,
                            "created_at": created_at,
                            "updated_at": updated_at,
                            "args": result["key"]["kw"]["args"],
                            "runner_type": runner_type,
                            "runner": result["key"]["kw"]["runner"],
                            "hooks": result["hooks"],
                            "sla": result["key"]["kw"]["sla"],
                            "sla_results": {"sla": result["sla"]},
                            "pass_sla": pass_sla,
                            "contexts": result["key"]["kw"]["context"],
                            "contexts_results": [],
                            "data": sorted(result["result"],
                                           key=lambda x: x["timestamp"]),
                            "statistics": {
                                "durations": durations_stat.to_dict()},
                            }
                task["subtasks"].append(
                    {"title": "A SubTask",
                     "description": "",
                     "workloads": [workload]})
            return [task]
        elif isinstance(tasks_results, dict) and "tasks" in tasks_results:
            for task_result in tasks_results["tasks"]:
                try:
                    jsonschema.validate(task_result,
                                        api.task.TASK_SCHEMA)
                except jsonschema.ValidationError as e:
                    msg = str(e)
                    raise exceptions.RallyException(
                        "ERROR: Invalid task result format\n\n\t%s" % msg)
                task_result.setdefault("env_name", "n/a")
                task_result.setdefault("env_uuid", "n/a")
                for subtask in task_result["subtasks"]:
                    for workload in subtask["workloads"]:
                        workload.setdefault("contexts_results", [])
                        workload["runner_type"], workload["runner"] = list(
                            workload["runner"].items())[0]
                        workload["name"], workload["args"] = list(
                            workload.pop("scenario").items())[0]

            return tasks_results["tasks"]
        else:
            raise FailedToLoadResults(
                source=task_id, msg="Wrong format")

    @cliutils.args("--out", metavar="<path>",
                   type=str, dest="out", required=False,
                   help="Path to output file.")
    @cliutils.args("--open", dest="open_it", action="store_true",
                   help="Open the output in a browser.")
    @cliutils.args("--tasks", dest="tasks", nargs="+",
                   help="UUIDs of tasks, or JSON files with task results")
    @cliutils.args("--html-static", dest="out_format",
                   action="store_const", const="trends-html-static")
    @cliutils.suppress_warnings
    def trends(self, api, *args, **kwargs):
        """Generate workloads trends HTML report."""
        tasks = kwargs.get("tasks", []) or list(args)
        if not tasks:
            print("ERROR: At least one task must be specified",
                  file=sys.stderr)
            return 1

        self.export(api, tasks=tasks,
                    output_type=kwargs.get("out_format", "trends-html"),
                    output_dest=kwargs.get("out"),
                    open_it=kwargs.get("open_it", False))

    @cliutils.args("--out", metavar="<path>",
                   type=str, dest="out", required=False,
                   help="Report destination. Can be a path to a file (in case"
                        " of HTML, HTML-STATIC, etc. types) to save the"
                        " report to or a connection string.")
    @cliutils.args("--open", dest="open_it", action="store_true",
                   help="Open the output in a browser.")
    @cliutils.args("--html", dest="out_format",
                   action="store_const", const="html")
    @cliutils.args("--html-static", dest="out_format",
                   action="store_const", const="html-static")
    @cliutils.args("--json", dest="out_format",
                   action="store_const", const="json")
    @cliutils.args("--uuid", dest="tasks", nargs="+", type=str,
                   help="UUIDs of tasks or json reports of tasks")
    @cliutils.args("--deployment", dest="deployment", type=str,
                   help="Report all tasks with defined deployment",
                   required=False)
    @envutils.default_from_global("tasks", envutils.ENV_TASK, "uuid")
    @cliutils.suppress_warnings
    def report(self, api, tasks=None, out=None,
               open_it=False, out_format="html", deployment=None):
        """Generate a report for the specified task(s)."""
        self.export(api, tasks=tasks,
                    output_type=out_format,
                    output_dest=out,
                    open_it=open_it,
                    deployment=deployment)

    @cliutils.args("--force", action="store_true", help="force delete")
    @cliutils.args("--uuid", type=str, dest="task_id", nargs="*",
                   metavar="<task-id>",
                   help="UUID of task or a list of task UUIDs.")
    @envutils.with_default_task_id
    def delete(self, api, task_id=None, force=False):
        """Delete task and its results."""

        def _delete_single_task(tid, force):
            try:
                api.task.delete(task_uuid=tid, force=force)
                print("Successfully deleted task `%s`" % tid)
            except exceptions.DBConflict as e:
                print(e)
                print("Use '--force' option to delete the task with vague "
                      "state.")

        if isinstance(task_id, list):
            for tid in task_id:
                _delete_single_task(tid, force)
        else:
            _delete_single_task(task_id, force)

    @cliutils.args("--uuid", type=str, dest="task_id", help="UUID of task.")
    @cliutils.args("--json", dest="tojson",
                   action="store_true",
                   help="Output in JSON format.")
    @envutils.with_default_task_id
    def sla_check(self, api, task_id=None, tojson=False):
        """Display SLA check results table."""

        task = api.task.get(task_id=task_id, detailed=True)
        failed_criteria = 0
        data = []
        STATUS_PASS = "PASS"
        STATUS_FAIL = "FAIL"
        for workload in itertools.chain(
                *[s["workloads"] for s in task["subtasks"]]):
            for sla in sorted(workload["sla_results"].get("sla", []),
                              key=lambda x: x["criterion"]):
                success = sla.pop("success")
                sla["status"] = success and STATUS_PASS or STATUS_FAIL
                sla["benchmark"] = workload["name"]
                sla["pos"] = workload["position"]
                failed_criteria += int(not success)
                data.append(sla if tojson else rutils.Struct(**sla))
        if tojson:
            print(json.dumps(data, sort_keys=False))
        else:
            cliutils.print_list(data, ("benchmark", "pos", "criterion",
                                       "status", "detail"))
        if not data:
            return 2
        return failed_criteria

    @cliutils.args("--uuid", type=str, dest="task_id",
                   help="UUID of the task")
    def use(self, api, task_id):
        """Set active task."""

        print("Using task: %s" % task_id)
        api.task.get(task_id=task_id)
        fileutils.update_globals_file("RALLY_TASK", task_id)

    @cliutils.args("--uuid", dest="tasks", nargs="+", type=str,
                   help="UUIDs of tasks or json reports of tasks")
    @cliutils.args("--type", dest="output_type", type=str,
                   required=True,
                   help="Report type. Out-of-the-box "
                        "types: JSON, HTML, HTML-Static, Elastic, JUnit-XML. "
                        "HINT: You can list all types, executing "
                        "`rally plugin list --plugin-base TaskExporter` "
                        "command.")
    @cliutils.args("--to", dest="output_dest", type=str,
                   metavar="<dest>", required=False,
                   help="Report destination. Can be a path to a file (in case"
                        " of JSON, HTML, HTML-Static, JUnit-XML, Elastic etc. "
                        "types) to save the report to or a connection string."
                        " It depends on the report type."
                   )
    @cliutils.args("--deployment", dest="deployment", type=str,
                   help="Report all tasks with defined deployment",
                   required=False)
    @envutils.default_from_global("tasks", envutils.ENV_TASK, "uuid")
    @plugins.ensure_plugins_are_loaded
    def export(self, api, tasks=None, output_type=None, output_dest=None,
               open_it=False, deployment=None):
        """Export task results to the custom task's exporting system."""

        if deployment is not None:
            tasks = api.task.list(deployment=deployment, uuids_only=True)
            tasks = [task["uuid"] for task in tasks]
        else:
            tasks = isinstance(tasks, list) and tasks or [tasks]

        exported_tasks = []
        for task_file_or_uuid in tasks:
            if os.path.exists(os.path.expanduser(task_file_or_uuid)):
                exported_tasks.extend(
                    self._load_task_results_file(api, task_file_or_uuid)
                )
            else:
                exported_tasks.append(task_file_or_uuid)

        report = api.task.export(tasks=exported_tasks,
                                 output_type=output_type,
                                 output_dest=output_dest)
        if "files" in report:
            for path in report["files"]:
                output_file = os.path.expanduser(path)
                with open(output_file, "w+") as f:
                    f.write(report["files"][path])
                if open_it:
                    if "open" in report:
                        webbrowser.open_new_tab(report["open"])

        if "print" in report:
            print(report["print"])

    @staticmethod
    def _print_task_errors(task_id, task_errors):
        print(cliutils.make_header("Task %s has %d error(s)" %
                                   (task_id, len(task_errors))))
        for err_data in task_errors:
            print(*err_data, sep="\n")
            print("-" * 80)

    @staticmethod
    def _format_task_error(data):
        error_type = "Unknown type"
        error_message = "Rally hasn't caught anything yet"
        error_traceback = "No traceback available."
        try:
            error_type = data["error"][0]
            error_message = data["error"][1]
            error_traceback = data["error"][2]
        except IndexError:
            pass
        return ("%(error_type)s: %(error_message)s\n" %
                {"error_type": error_type, "error_message": error_message},
                error_traceback)

    @cliutils.args("--file", dest="task_file", type=str, metavar="<path>",
                   required=True, help="JSON file with task results")
    @cliutils.args("--deployment", dest="deployment", type=str,
                   metavar="<uuid>", required=False,
                   help="UUID or name of a deployment.")
    @cliutils.args("--tag", nargs="+", dest="tags", type=str, required=False,
                   help="Mark the task with a tag or a few tags.")
    @envutils.with_default_deployment(cli_arg_name="deployment")
    @cliutils.alias("import")
    @cliutils.suppress_warnings
    def import_results(self, api, deployment=None, task_file=None, tags=None):
        """Import json results of a test into rally database"""

        if os.path.exists(os.path.expanduser(task_file)):
            tasks_results = self._load_task_results_file(api, task_file)
            for task_results in tasks_results:
                task = api.task.import_results(deployment=deployment,
                                               task_results=task_results,
                                               tags=tags)
                print("Task UUID: %s." % task["uuid"])
        else:
            print("ERROR: Invalid file name passed: %s" % task_file,
                  file=sys.stderr)
            return 1



# Copyright 2015: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import functools

from rally.common import logging
from rally.common import utils

LOG = logging.getLogger(__name__)


class ActionTimerMixin(object):

    def __init__(self):
        self._atomic_actions = []

    def atomic_actions(self):
        """Returns the content of each atomic action."""
        return self._atomic_actions

    def reset_atomic_actions(self):
        """Clean all atomic action data."""
        self._atomic_actions = []


class ActionTimer(utils.Timer):
    """A class to measure the duration of atomic operations

    This would simplify the way measure atomic operation duration
    in certain cases. For example, if we want to get the duration
    for each operation which runs in an iteration
    for i in range(repetitions):
        with atomic.ActionTimer(instance_of_action_timer, "name_of_action"):
            self.clients(<client>).<operation>
    """

    def __init__(self, instance, name):
        """Create a new instance of the AtomicAction.

        :param instance: instance of subclass of ActionTimerMixin
        :param name: name of the ActionBuilder
        """
        super(ActionTimer, self).__init__()
        self.instance = instance
        self.name = name
        self._root = self._find_parent(self.instance._atomic_actions)
        self.atomic_action = {"name": self.name,
                              "children": [],
                              "started_at": None}
        self._root.append(self.atomic_action)

    def _find_parent(self, atomic_actions):
        while atomic_actions and "finished_at" not in atomic_actions[-1]:
            atomic_actions = atomic_actions[-1]["children"]
        return atomic_actions

    def __enter__(self):
        super(ActionTimer, self).__enter__()
        self.atomic_action["started_at"] = self.start

    def __exit__(self, type_, value, tb):
        super(ActionTimer, self).__exit__(type_, value, tb)
        self.atomic_action["finished_at"] = self.finish
        if type_:
            self.atomic_action["failed"] = True


def action_timer(name):
    """Provide measure of execution time.

    Decorates methods of the Scenario class.
    This provides duration in seconds of each atomic action.
    """
    def wrap(func):
        @functools.wraps(func)
        def func_atomic_actions(self, *args, **kwargs):
            with ActionTimer(self, name):
                f = func(self, *args, **kwargs)
            return f
        return func_atomic_actions
    return wrap


def merge_atomic_actions(atomic_actions, root=None, depth=0,
                         depth_of_processing=2):
    """Merge duplicates of atomic actions into one atomic action.

    :param atomic_actions: a list with atomic action
    :param root: an ordered dict to save atomics to (leave it with a default
        value, since the primary use case of that parameter is processing inner
        atomic actions)
    :param depth: current level of processing inner atomic actions
    :param depth_of_processing: the depth of processing of inner atomic actions
        (defaults to 2)
    """
    p_atomics = collections.OrderedDict() if root is None else root
    for action in atomic_actions:
        if action["name"] not in p_atomics:
            p_atomics[action["name"]] = {
                "duration": 0,
                "count": 0,
                "children": collections.OrderedDict()}
        duration = action["finished_at"] - action["started_at"]
        p_atomics[action["name"]]["duration"] += duration
        p_atomics[action["name"]]["count"] += 1
        if action.get("failed"):
            p_atomics[action["name"]]["failed"] = True
        if action["children"] and depth < depth_of_processing:
            merge_atomic_actions(
                action["children"],
                root=p_atomics[action["name"]]["children"],
                depth=depth + 1)

    return p_atomics



# Copyright 2015: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import importlib
import os
import pkg_resources
import pkgutil
import sys

import rally
from rally.common import logging

LOG = logging.getLogger(__name__)


def itersubclasses(cls, seen=None):
    """Generator over all subclasses of a given class in depth first order.

    NOTE: Use 'seen' to exclude cls which was reduplicated found, because
    cls maybe has multiple super classes of the same plugin.
    """

    seen = seen or set()
    try:
        subs = cls.__subclasses__()
    except TypeError:   # fails only when cls is type
        subs = cls.__subclasses__(cls)
    for sub in subs:
        if sub not in seen:
            seen.add(sub)
            yield sub
            for sub in itersubclasses(sub, seen):
                yield sub


def import_modules_from_package(package):
    """Import modules from package and append into sys.modules

    :param package: Full package name. For example: rally.plugins.openstack
    """
    path = [os.path.dirname(rally.__file__), ".."] + package.split(".")
    path = os.path.join(*path)
    for root, dirs, files in os.walk(path):
        for filename in files:
            if filename.startswith("__") or not filename.endswith(".py"):
                continue
            new_package = ".".join(root.split(os.sep)).split("....")[1]
            module_name = "%s.%s" % (new_package, filename[:-3])
            if module_name not in sys.modules:
                sys.modules[module_name] = importlib.import_module(module_name)


def find_packages_by_entry_point():
    """Find all packages with rally_plugins entry-point"""
    loaded_packages = []

    for package in pkg_resources.working_set:
        entry_map = package.get_entry_map("rally_plugins")
        if not entry_map:
            # this package doesn't have rally_plugins entry-point
            continue

        package_info = {}

        if "path" in entry_map:
            ep = entry_map["path"]
            package_info["plugins_path"] = ep.module_name
        if "options" in entry_map:
            ep = entry_map["options"]
            package_info["options"] = "%s:%s" % (
                ep.module_name,
                ep.attrs[0] if ep.attrs else "list_opts",
            )

        if package_info:
            package_info.update(
                name=package.project_name,
                version=package.version)
            loaded_packages.append(package_info)
    return loaded_packages


def import_modules_by_entry_point(_packages=None):
    """Import plugins by entry-point 'rally_plugins'."""
    loaded_packages = _packages or find_packages_by_entry_point()

    for package in loaded_packages:
        if "plugins_path" in package:
            em = pkg_resources.get_entry_map(package["name"])
            ep = em["rally_plugins"]["path"]
            try:
                m = ep.load()
                if hasattr(m, "__path__"):
                    path = pkgutil.extend_path(m.__path__, m.__name__)
                else:
                    path = [m.__file__]
                prefix = m.__name__ + "."
                for loader, name, _is_pkg in pkgutil.walk_packages(
                        path, prefix=prefix):
                    sys.modules[name] = importlib.import_module(name)
            except Exception as e:
                msg = ("\t Failed to load plugins from module '%(module)s' "
                       "(package: '%(package)s')" %
                       {"module": ep.module_name,
                        "package": "%s %s" % (package["name"],
                                              package["version"])})
                if logging.is_debug():
                    LOG.exception(msg)
                else:
                    LOG.warning(msg + (": %s" % str(e)))
    return loaded_packages


_loaded_modules = []


def load_plugins(dir_or_file, depth=0):
    if os.path.isdir(dir_or_file):
        directory = dir_or_file
        LOG.info("Loading plugins from directories %s/*" %
                 directory.rstrip("/"))
        for root, _dirs, files in os.walk(directory, followlinks=True):
            if root not in sys.path:
                # this hack is required to support relative imports
                sys.path.append(root)

            for plugin in files:
                load_plugins(os.path.join(root, plugin), depth=1)

    elif os.path.isfile(dir_or_file) and dir_or_file.endswith(".py"):
        plugin_file = dir_or_file
        msg = "Loading plugins from file %s" % plugin_file
        if depth:
            msg = "\t" + msg
        LOG.info(msg)
        module_name = os.path.splitext(plugin_file.split("/")[-1])[0]
        try:
            spec = importlib.util.spec_from_file_location(
                module_name, plugin_file)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)

        except Exception as e:
            msg = "\t Failed to load module with plugins %s" % plugin_file
            if logging.is_debug():
                LOG.exception(msg)
            else:
                LOG.warning("%(msg)s: %(e)s" % {"msg": msg, "e": e})
            return
        _loaded_modules.append(module)
        LOG.info("\t Loaded module with plugins: %s" % module_name)



# Copyright 2014: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
import mock

from rally.common.plugin import plugin
from rally.task import sla
from tests.unit import test


class SLABaseForTestCase(test.TestCase):

    def setUp(self):
        super(SLABaseForTestCase, self).setUp()

        @plugin.configure(name="test_criterion")
        class FakeCriterion(sla.SLA):
            CONFIG_SCHEMA = {"type": "integer"}

            def add_iteration(self, iteration):
                self.success = self.criterion_value == iteration
                return self.success

            def merge(self, other):
                raise NotImplementedError()

            def details(self):
                return "detail"

        self.FakeCriterion = FakeCriterion
        self.addCleanup(FakeCriterion.unregister)


@ddt.ddt
class SLACheckerTestCase(SLABaseForTestCase):

    def test_add_iteration_and_results(self):
        sla_checker = sla.SLAChecker({"sla": {"test_criterion": 42}})

        iteration = {"key": {"name": "fake", "pos": 0}, "data": 42}
        self.assertTrue(sla_checker.add_iteration(iteration["data"]))
        expected_result = [{"criterion": "test_criterion",
                            "detail": "detail",
                            "success": True}]
        self.assertEqual(expected_result, sla_checker.results())

        iteration["data"] = 43
        self.assertFalse(sla_checker.add_iteration(iteration["data"]))
        expected_result = [{"criterion": "test_criterion",
                            "detail": "detail",
                            "success": False}]
        self.assertEqual(expected_result, sla_checker.results())

    def test_set_unexpected_failure(self):
        exc = "error;("
        sla_checker = sla.SLAChecker({"sla": {}})
        self.assertEqual([], sla_checker.results())
        sla_checker.set_unexpected_failure(exc)
        self.assertEqual([{"criterion": "something_went_wrong",
                           "success": False,
                           "detail": "Unexpected error: %s" % exc}],
                         sla_checker.results())

    def test_set_aborted_on_sla(self):
        sla_checker = sla.SLAChecker({"sla": {}})
        self.assertEqual([], sla_checker.results())
        sla_checker.set_aborted_on_sla()
        self.assertEqual(
            [{"criterion": "aborted_on_sla", "success": False,
              "detail": "Task was aborted due to SLA failure(s)."}],
            sla_checker.results())

    def test_set_aborted_manually(self):
        sla_checker = sla.SLAChecker({"sla": {}})
        self.assertEqual([], sla_checker.results())
        sla_checker.set_aborted_manually()
        self.assertEqual(
            [{"criterion": "aborted_manually", "success": False,
              "detail": "Task was aborted due to abort signal."}],
            sla_checker.results())

    def test__format_result(self):
        name = "some_name"
        success = True
        detail = "some details"
        self.assertEqual({"criterion": name,
                          "success": success,
                          "detail": detail},
                         sla._format_result(name, success, detail))

    def test__validate_config_positive(self):
        sla_checker = sla.SLAChecker({"sla": {}})
        another_sla_checker = sla.SLAChecker({"sla": {}})
        sla_checker._validate_config(another_sla_checker)

    def test__validate_config_negative(self):
        sla_checker = sla.SLAChecker({"sla": {}})
        another_sla_checker = sla.SLAChecker({"sla": {"test_criterion": 42}})
        self.assertRaises(TypeError, sla_checker._validate_config,
                          another_sla_checker)

    def test__validate_sla_types(self):
        sla_checker = sla.SLAChecker({"sla": {}})
        mock_sla1 = mock.MagicMock()
        mock_sla2 = mock.MagicMock()
        sla_checker.sla_criteria = [mock_sla1, mock_sla2]

        another_sla_checker = sla.SLAChecker({"sla": {}})
        mock_sla3 = mock.MagicMock()
        mock_sla4 = mock.MagicMock()
        another_sla_checker.sla_criteria = [mock_sla3, mock_sla4]

        sla_checker._validate_sla_types(another_sla_checker)

        mock_sla1.assert_has_calls([
            mock.call.validate_type(mock_sla3)
        ])

        mock_sla1.validate_type.assert_called_once_with(mock_sla3)
        mock_sla2.validate_type.assert_called_once_with(mock_sla4)

    @ddt.data({"merge_result1": True, "merge_result2": True,
               "result": True},
              {"merge_result1": True, "merge_result2": False,
               "result": False},
              {"merge_result1": False, "merge_result2": False,
               "result": False})
    @ddt.unpack
    def test_merge(self, merge_result1, merge_result2, result):
        sla_checker = sla.SLAChecker({"sla": {}})
        mock_sla1 = mock.MagicMock()
        mock_sla2 = mock.MagicMock()
        sla_checker.sla_criteria = [mock_sla1, mock_sla2]

        mock_sla1.merge.return_value = merge_result1
        mock_sla2.merge.return_value = merge_result2

        another_sla_checker = sla.SLAChecker({"sla": {}})
        mock_sla3 = mock.MagicMock()
        mock_sla4 = mock.MagicMock()
        another_sla_checker.sla_criteria = [mock_sla3, mock_sla4]

        sla_checker._validate_config = mock.MagicMock()
        sla_checker._validate_sla_types = mock.MagicMock()

        self.assertEqual(result, sla_checker.merge(another_sla_checker))
        mock_sla1.merge.assert_called_once_with(mock_sla3)
        mock_sla2.merge.assert_called_once_with(mock_sla4)


@ddt.ddt
class SLATestCase(SLABaseForTestCase):
    def test_validate_type_positive(self):
        sla1 = self.FakeCriterion(0)
        sla2 = self.FakeCriterion(0)
        sla1.validate_type(sla2)

    def test_validate_type_negative(self):
        sla1 = self.FakeCriterion(0)

        class AnotherFakeCriterion(self.FakeCriterion):
            pass

        sla2 = AnotherFakeCriterion(0)
        self.assertRaises(TypeError, sla1.validate_type, sla2)

    @ddt.data((10, True),
              ({}, False))
    @ddt.unpack
    def test_validate(self, config, valid):
        results = sla.SLA.validate(
            "test_criterion", None, None, config)
        if valid:
            self.assertEqual([], results)
        else:
            self.assertEqual(1, len(results))



# Copyright 2017: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from docutils import nodes
import json

from oslo_utils import importutils


def include_var(name, rawtext, text, lineno, inliner, options=None,
                content=None):
    """include variable

    :param name: The local name of the interpreted role, the role name
                 actually used in the document.
    :param rawtext: A string containing the enitre interpreted text input,
                    including the role and markup. Return it as a problematic
                    node linked to a system message if a problem is
                    encountered.
    :param text: The interpreted text content.
    :param lineno: The line number where the interpreted text begins.
    :param inliner: The docutils.parsers.rst.states.Inliner object that
                    called include_var. It contains the several attributes
                    useful for error reporting and document tree access.
    :param options: A dictionary of directive options for customization
                    (from the 'role' directive), to be interpreted by the
                    role function. Used for additional attributes for the
                    generated elements and other functionality.
    :param content: A list of strings, the directive content for
                    customization (from the 'role' directive). To be
                    interpreted by the role function.
    :return:
    """
    obj = importutils.import_class(text)
    if isinstance(obj, (tuple, list)):
        obj = ", ".join(obj)
    elif isinstance(obj, dict):
        obj = json.dumps(dict, indent=4)
    else:
        obj = str(obj)
    return [nodes.inline("", obj)], []


def setup(app):
    app.add_role("include-var", include_var)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

import ddt
import json
import mock

from rally import exceptions
from rally.plugins.common.exporters.elastic import exporter as elastic
from tests.unit import test


PATH = "rally.plugins.common.exporters.elastic.exporter"


class ValidatorTestCase(test.TestCase):
    @mock.patch("%s.client.ElasticSearchClient" % PATH)
    def test_validate(self, mock_elastic_search_client):
        validator = elastic.Validator()
        client = mock_elastic_search_client.return_value

        validator.validate({}, {}, None, {"destination": "/home/foo"})
        self.assertFalse(mock_elastic_search_client.called)

        client.version.return_value = "2.5.1"
        validator.validate({}, {}, None, {"destination": None})

        client.version.return_value = "5.6.2"
        validator.validate({}, {}, None, {"destination": None})

        client.version.return_value = "1.1.1"
        e = self.assertRaises(
            elastic.validation.ValidationError,
            validator.validate, {}, {}, None, {"destination": None})
        self.assertEqual("The unsupported version detected 1.1.1.",
                         e.message)

        exp_e = exceptions.RallyException("foo")
        client.version.side_effect = exp_e
        actual_e = self.assertRaises(
            elastic.validation.ValidationError,
            validator.validate, {}, {}, None, {"destination": None})
        self.assertEqual(exp_e.format_message(), actual_e.message)


def get_tasks_results():
    task_uuid = "2fa4f5ff-7d23-4bb0-9b1f-8ee235f7f1c8"
    subtask_uuid = "35166362-0b11-4e74-929d-6988377e2da2"
    return [{
        "id": 1,
        "uuid": task_uuid,
        "deployment_uuid": "deployment-uuu-iii-iii-ddd",
        "env_uuid": "deployment-uuu-iii-iii-ddd",
        "env_name": "env-name",
        "title": "foo",
        "description": "bar",
        "status": "ok",
        "pass_sla": "yup",
        "task_duration": "dur",
        "tags": ["tag-1", "tag-2"],
        "subtasks": [{
            "task_uuid": task_uuid,
            "subtask_uuid": subtask_uuid,
            "workloads": [{
                "id": 3,
                "position": 0,
                "uuid": "4dcd88a5-164b-4431-8b44-3979868116dd",
                "task_uuid": task_uuid,
                "subtask_uuid": subtask_uuid,
                "name": "CinderVolumes.list_volumes",
                "args": {"key1": "value1"},
                "description": "List all volumes.",
                "runner_type": "constant",
                "runner": {"type": "constant",
                           "times": 3},
                "sla": {},
                "contexts": {"users@openstack": {"tenants": 2}},
                "created_at": "2017-07-28T23:35:46",
                "updated_at": "2017-07-28T23:37:55",
                "start_time": 1501284950.371992,
                "failed_iteration_count": 2,
                "load_duration": 97.82577991485596,
                "full_duration": 127.59103488922119,
                "pass_sla": False,
                "sla_results": {
                    "sla": [{"criterion": "JoGo",
                             "success": False,
                             "detail": "because i do not like you"}]
                },
                "statistics": {
                    "durations": {
                        "total": {
                            "data": {"success": "80.0%"}
                        }
                    }
                },
                "data": [
                    # iteration where the unwrapped action failed
                    {"timestamp": 1501284950.371992,
                     "error": ["ERROR!"],
                     "duration": 10.096552848815918,
                     "idle_duration": 0,
                     "atomic_actions": [
                         {"finished_at": 1501284961.468537,
                          "started_at": 1501284950.372052,
                          "name": "cinder.list_volumes",
                          "children": []}]},
                    # iteration where the known action failed
                    {"timestamp": 1501284950.371992,
                     "error": ["ERROR!"],
                     "duration": 10.096552848815918,
                     "idle_duration": 0,
                     "atomic_actions": [
                         {"finished_at": 1501284961.468537,
                          "started_at": 1501284950.372052,
                          "name": "cinder.list_volumes",
                          "failed": True,
                          "children": []}]}
                ]}]
        }]
    }]


@ddt.ddt
class ElasticSearchExporterTestCase(test.TestCase):
    def setUp(self):
        super(ElasticSearchExporterTestCase, self).setUp()
        self.patcher = mock.patch.object(elastic.client, "ElasticSearchClient")
        self.es_cls = self.patcher.start()
        self.addCleanup(self.patcher.stop)

    def test_init(self):
        exporter = elastic.ElasticSearchExporter([], "http://example.com")
        self.assertTrue(exporter._remote)
        self.assertEqual(self.es_cls.return_value,
                         getattr(exporter, "_client"))

        exporter = elastic.ElasticSearchExporter([], None)
        self.assertTrue(exporter._remote)
        self.assertEqual(self.es_cls.return_value,
                         getattr(exporter, "_client"))

        exporter = elastic.ElasticSearchExporter([], "/foo/bar")
        self.assertFalse(exporter._remote)
        self.assertIsNone(getattr(exporter, "_client", None))

    @ddt.data(None, "/home/bar", "https://example.com")
    def test__add_index(self, destination):

        index = "foo"
        doc_type = "bar"
        body = {
            "key1": "value1",
            "key2": "value2"
        }
        doc_id = "2fa4f5ff-7d23-4bb0-9b1f-8ee235f7f1c8"

        exporter = elastic.ElasticSearchExporter([], destination)

        exporter._add_index(index=index,
                            body=body,
                            doc_id=doc_id,
                            doc_type=doc_type)

        self.assertEqual(2, len(exporter._report))
        self.assertEqual({"index": {"_index": index,
                                    "_type": doc_type,
                                    "_id": doc_id}},
                         json.loads(exporter._report[0]))

    @ddt.data(True, False)
    @mock.patch("%s.ElasticSearchExporter._add_index" % PATH)
    def test__process_atomic_actions(self, known_fail, mock__add_index):
        es_exporter = elastic.ElasticSearchExporter({}, None)

        itr_data = {"id": "foo_bar_uuid",
                    "error": ["I was forced to fail. Sorry"],
                    "timestamp": 1, "duration": 2, "idle_duration": 1}
        workload = {
            "scenario_cfg": ["key1=value1"],
            "runner_name": "foo",
            "runner_cfg": ["times=3"],
            "contexts": ["users@openstack.tenants=2"],
            "deployment_uuid": "dep_uuid", "deployment_name": "dep_name"}

        atomic_actions = [
            {"name": "do_something",
             "started_at": 1, "finished_at": 2,
             "children": []},
            {"name": "fail_something",
             "started_at": 3,
             "finished_at": 4,
             "children": [
                 {"name": "rm -rf", "started_at": 3, "finished_at": 4,
                  "children": []}
             ]},
        ]

        if known_fail:
            atomic_actions[-1]["failed"] = True
            atomic_actions[-1]["children"][-1]["failed"] = True

        es_exporter._process_atomic_actions(
            atomic_actions=atomic_actions, itr=itr_data,
            workload_id="wid", workload=workload)

        expected_calls = [
            mock.call(
                "rally_atomic_action_data_v1",
                {
                    "deployment_uuid": "dep_uuid",
                    "deployment_name": "dep_name",
                    "action_name": "do_something",
                    "scenario_cfg": ["key1=value1"],
                    "contexts": ["users@openstack.tenants=2"],
                    "runner_name": "foo",
                    "runner_cfg": ["times=3"],
                    "started_at": "1970-01-01T00:00:01",
                    "finished_at": "1970-01-01T00:00:02",
                    "duration": 1,
                    "success": True,
                    "error": None,
                    "parent": None,
                    "workload_uuid": "wid"},
                doc_id="foo_bar_uuid_action_do_something_0"),
            mock.call(
                "rally_atomic_action_data_v1",
                {
                    "deployment_uuid": "dep_uuid",
                    "deployment_name": "dep_name",
                    "action_name": "fail_something",
                    "scenario_cfg": ["key1=value1"],
                    "contexts": ["users@openstack.tenants=2"],
                    "runner_name": "foo",
                    "runner_cfg": ["times=3"],
                    "started_at": "1970-01-01T00:00:03",
                    "finished_at": "1970-01-01T00:00:04",
                    "duration": 1,
                    "success": not known_fail,
                    "error": itr_data["error"] if known_fail else None,
                    "parent": None,
                    "workload_uuid": "wid"},
                doc_id="foo_bar_uuid_action_fail_something_0"),
            mock.call(
                "rally_atomic_action_data_v1",
                {
                    "deployment_uuid": "dep_uuid",
                    "deployment_name": "dep_name",
                    "action_name": "rm -rf",
                    "scenario_cfg": ["key1=value1"],
                    "contexts": ["users@openstack.tenants=2"],
                    "runner_name": "foo",
                    "runner_cfg": ["times=3"],
                    "started_at": "1970-01-01T00:00:03",
                    "finished_at": "1970-01-01T00:00:04",
                    "duration": 1,
                    "success": not known_fail,
                    "error": itr_data["error"] if known_fail else None,
                    "parent": "foo_bar_uuid_action_fail_something_0",
                    "workload_uuid": "wid"},
                doc_id="foo_bar_uuid_action_rm -rf_0")]

        if not known_fail:
            expected_calls.append(mock.call(
                "rally_atomic_action_data_v1",
                {
                    "deployment_uuid": "dep_uuid",
                    "deployment_name": "dep_name",
                    "action_name": "no-name-action",
                    "scenario_cfg": ["key1=value1"],
                    "contexts": ["users@openstack.tenants=2"],
                    "runner_name": "foo",
                    "runner_cfg": ["times=3"],
                    "started_at": "1970-01-01T00:00:04",
                    "finished_at": "1970-01-01T00:00:04",
                    "duration": 0,
                    "success": False,
                    "error": itr_data["error"],
                    "parent": None,
                    "workload_uuid": "wid"},
                doc_id="foo_bar_uuid_action_no-name-action_0"))

        self.assertEqual(expected_calls, mock__add_index.call_args_list)

    def test_generate_fails_on_doc_exists(self):
        destination = "http://example.com"
        client = self.es_cls.return_value
        client.check_document.side_effect = (False, True)

        tasks = get_tasks_results()
        second_task = copy.deepcopy(tasks[-1])
        second_task["subtasks"] = []
        tasks.append(second_task)

        exporter = elastic.ElasticSearchExporter(tasks, destination)

        e = self.assertRaises(exceptions.RallyException, exporter.generate)
        self.assertIn("Failed to push the task %s" % tasks[0]["uuid"],
                      e.format_message())

    def test__ensure_indices(self):
        es = mock.MagicMock()
        exporter = elastic.ElasticSearchExporter([], None)
        exporter._client = es

        # case #1: everything exists
        es.list_indices.return_value = [exporter.WORKLOAD_INDEX,
                                        exporter.TASK_INDEX,
                                        exporter.AA_INDEX]
        exporter._ensure_indices()

        self.assertFalse(es.create_index.called)
        es.list_indices.assert_called_once_with()

        # case #2: some indices exist
        es.list_indices.reset_mock()
        es.list_indices.return_value = [exporter.TASK_INDEX, exporter.AA_INDEX]

        exporter._ensure_indices()

        es.list_indices.assert_called_once_with()
        es.create_index.assert_called_once_with(
            exporter.WORKLOAD_INDEX, doc_type="data",
            properties=exporter.INDEX_SCHEMAS[exporter.WORKLOAD_INDEX]
        )

        # case #3: none of indices exists
        es.list_indices.reset_mock()
        es.create_index.reset_mock()
        es.list_indices.return_value = []

        exporter._ensure_indices()

        es.list_indices.assert_called_once_with()
        self.assertEqual(3, es.create_index.call_count)

    @ddt.data(True, False)
    def test_generate(self, remote):
        if remote:
            destination = "http://example.com"
            client = self.es_cls.return_value
            client.check_document.return_value = False
        else:
            destination = "/home/bar.txt"

        tasks = get_tasks_results()
        second_task = copy.deepcopy(tasks[-1])
        second_task["subtasks"] = []
        tasks.append(second_task)

        exporter = elastic.ElasticSearchExporter(tasks, destination)

        result = exporter.generate()

        if remote:
            self.assertEqual(
                [mock.call("rally_task_data_v1", second_task["uuid"]),
                 mock.call("rally_task_data_v1", second_task["uuid"])],
                client.check_document.call_args_list
            )
            client.push_documents.assert_called_once_with(exporter._report)
            client.list_indices.assert_called_once_with()
            self.assertEqual(3, client.create_index.call_count)
        else:
            self.assertEqual({"files", "open"}, set(result.keys()))
            self.assertEqual("file://%s" % destination, result["open"])
            self.assertEqual({destination}, set(result["files"].keys()))

            data = result["files"][destination].split("\n")
            # the should be always empty line in the end
            self.assertEqual("", data[-1])

        data = [json.loads(l) for l in exporter._report]
        self.assertIsInstance(data, list)
        expected = [
            {
                "index": {"_id": "2fa4f5ff-7d23-4bb0-9b1f-8ee235f7f1c8",
                          "_index": "rally_task_data_v1",
                          "_type": "data"}
            },
            {
                "title": "foo",
                "description": "bar",
                "deployment_uuid": "deployment-uuu-iii-iii-ddd",
                "deployment_name": "env-name",
                "status": "ok",
                "pass_sla": "yup",
                "task_uuid": "2fa4f5ff-7d23-4bb0-9b1f-8ee235f7f1c8",
                "tags": ["tag-1", "tag-2"]
            },
            {
                "index": {"_id": "4dcd88a5-164b-4431-8b44-3979868116dd",
                          "_index": "rally_workload_data_v1",
                          "_type": "data"}
            },
            {
                "deployment_uuid": "deployment-uuu-iii-iii-ddd",
                "deployment_name": "env-name",
                "task_uuid": "2fa4f5ff-7d23-4bb0-9b1f-8ee235f7f1c8",
                "subtask_uuid": "35166362-0b11-4e74-929d-6988377e2da2",
                "scenario_name": "CinderVolumes.list_volumes",
                "description": "List all volumes.",
                "scenario_cfg": ["key1=value1"],
                "contexts": ["users@openstack.tenants=2"],
                "runner_name": "constant",
                "runner_cfg": ["times=3", "type=constant"],
                "full_duration": 127.59103488922119,
                "load_duration": 97.82577991485596,
                "started_at": "2017-07-28T23:35:50",
                "pass_sla": False,
                "success_rate": 0.8,
                "sla_details": ["because i do not like you"]
            },
            {
                "index": {
                    "_id": "4dcd88a5-164b-4431-8b44-3979868116dd_iter_1_action"
                           "_cinder.list_volumes_0",
                    "_index": "rally_atomic_action_data_v1",
                    "_type": "data"}
            },
            {
                "deployment_uuid": "deployment-uuu-iii-iii-ddd",
                "deployment_name": "env-name",
                "action_name": "cinder.list_volumes",
                "started_at": "2017-07-28T23:35:50",
                "finished_at": "2017-07-28T23:36:01",
                "duration": 11.096485137939453,
                "contexts": ["users@openstack.tenants=2"],
                "error": None,
                "parent": None,
                "runner_name": "constant",
                "runner_cfg": ["times=3", "type=constant"],
                "scenario_cfg": ["key1=value1"],
                "success": True,
                "workload_uuid": "4dcd88a5-164b-4431-8b44-3979868116dd"
            },
            {
                "index": {
                    "_id": "4dcd88a5-164b-4431-8b44-3979868116dd_iter_1_action"
                           "_no-name-action_0",
                    "_index": "rally_atomic_action_data_v1",
                    "_type": "data"}
            },
            {
                "deployment_uuid": "deployment-uuu-iii-iii-ddd",
                "deployment_name": "env-name",
                "action_name": "no-name-action",
                "started_at": "2017-07-28T23:36:00",
                "finished_at": "2017-07-28T23:36:00",
                "duration": 0,
                "contexts": ["users@openstack.tenants=2"],
                "error": ["ERROR!"],
                "parent": None,
                "runner_name": "constant",
                "runner_cfg": ["times=3", "type=constant"],
                "scenario_cfg": ["key1=value1"],
                "success": False,
                "workload_uuid": "4dcd88a5-164b-4431-8b44-3979868116dd"
            },
            {
                "index": {
                    "_id": "4dcd88a5-164b-4431-8b44-3979868116dd_iter_2_action"
                           "_cinder.list_volumes_0",
                    "_index": "rally_atomic_action_data_v1",
                    "_type": "data"}
            },
            {
                "deployment_uuid": "deployment-uuu-iii-iii-ddd",
                "deployment_name": "env-name",
                "action_name": "cinder.list_volumes",
                "started_at": "2017-07-28T23:35:50",
                "finished_at": "2017-07-28T23:36:01",
                "duration": 11.096485137939453,
                "contexts": ["users@openstack.tenants=2"],
                "error": ["ERROR!"],
                "parent": None,
                "runner_name": "constant",
                "runner_cfg": ["times=3", "type=constant"],
                "scenario_cfg": ["key1=value1"],
                "success": False,
                "workload_uuid": "4dcd88a5-164b-4431-8b44-3979868116dd"
            },
            {
                "index": {"_id": "2fa4f5ff-7d23-4bb0-9b1f-8ee235f7f1c8",
                          "_index": "rally_task_data_v1",
                          "_type": "data"}
            },
            {
                "deployment_uuid": "deployment-uuu-iii-iii-ddd",
                "deployment_name": "env-name",
                "title": "foo",
                "description": "bar",
                "status": "ok",
                "pass_sla": "yup",
                "task_uuid": "2fa4f5ff-7d23-4bb0-9b1f-8ee235f7f1c8",
                "tags": ["tag-1", "tag-2"]}
        ]

        for i, line in enumerate(expected):
            if i == len(data):
                self.fail("The next line is missed: %s" % line)
            self.assertEqual(line, data[i], "Line #%s is wrong." % (i + 1))



# Copyright 2015: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from rally import consts
from rally import exceptions
from rally.task import context


@context.configure(name="dummy_context", order=750)
class DummyContext(context.Context):
    """Dummy context."""
    CONFIG_SCHEMA = {
        "type": "object",
        "$schema": consts.JSON_SCHEMA,
        "properties": {
            "fail_setup": {"type": "boolean"},
            "fail_cleanup": {"type": "boolean"}
        },
        "additionalProperties": False,
    }

    def setup(self):
        if self.config.get("fail_setup", False):
            raise exceptions.RallyException("Oops...setup is failed")

    def cleanup(self):
        if self.config.get("fail_cleanup", False):
            raise exceptions.RallyException("Oops...cleanup is failed")



# Copyright 2014: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import inspect
import itertools
import os
import traceback

import mock
import yaml

import rally
from rally import api
from rally.task import context
from rally.task import engine
from rally.task import scenario
from rally.task import task_cfg
from tests.unit import test

RALLY_PATH = os.path.dirname(os.path.dirname(rally.__file__))


class TaskSampleTestCase(test.TestCase):
    samples_path = os.path.join(RALLY_PATH, "samples", "tasks")

    def setUp(self):
        super(TaskSampleTestCase, self).setUp()
        if os.environ.get("TOX_ENV_NAME") == "cover":
            self.skipTest("There is no need to check samples in coverage job.")
        with mock.patch("rally.api.API.check_db_revision"):
            self.rapi = api.API()

    def iterate_samples(self, merge_pairs=True):
        """Iterates all task samples

        :param merge_pairs: Whether or not to return both json and yaml samples
            of one sample.
        """
        for dirname, dirnames, filenames in os.walk(self.samples_path):
            for filename in filenames:
                # NOTE(hughsaunders): Skip non config files
                # (bug https://bugs.launchpad.net/rally/+bug/1314369)
                if filename.endswith("json") or (
                        not merge_pairs and filename.endswith("yaml")):
                    yield os.path.join(dirname, filename)

    def test_check_missing_sla_section(self):
        failures = []
        for path in self.iterate_samples():
            if "tasks/scenarios" not in path:
                continue
            with open(path) as task_file:
                task_config = yaml.safe_load(
                    self.rapi.task.render_template(
                        task_template=task_file.read()))
                for workload in itertools.chain(*task_config.values()):
                    if not workload.get("sla", {}):
                        failures.append(path)
        if failures:
            self.fail("One or several workloads from the list of samples below"
                      " doesn't have SLA section: \n  %s" %
                      "\n  ".join(failures))

    def test_schema_is_valid(self):
        scenarios = set()

        for path in self.iterate_samples():
            with open(path) as task_file:
                try:
                    try:
                        task_config = yaml.safe_load(
                            self.rapi.task.render_template(
                                task_template=task_file.read()))
                    except Exception:
                        print(traceback.format_exc())
                        self.fail("Invalid JSON file: %s" % path)
                    eng = engine.TaskEngine(task_cfg.TaskConfig(task_config),
                                            mock.MagicMock(), mock.Mock())
                    eng.validate(only_syntax=True)
                except Exception:
                    print(traceback.format_exc())
                    self.fail("Invalid task file: %s" % path)
                else:
                    scenarios.update(task_config.keys())

        missing = set(s.get_name() for s in scenario.Scenario.get_all())
        missing -= scenarios
        # check missing scenario is not from plugin
        missing = [s for s in list(missing)
                   if scenario.Scenario.get(s).__module__.startswith("rally")]
        self.assertEqual(missing, [],
                         "These scenarios don't have samples: %s" % missing)

    def test_task_config_pairs(self):

        not_equal = []
        missed = []
        checked = []

        for path in self.iterate_samples(merge_pairs=False):
            if path.endswith(".json"):
                json_path = path
                yaml_path = json_path.replace(".json", ".yaml")
            else:
                yaml_path = path
                json_path = yaml_path.replace(".yaml", ".json")

            if json_path in checked:
                continue
            else:
                checked.append(json_path)

            if not os.path.exists(yaml_path):
                missed.append(yaml_path)
            elif not os.path.exists(json_path):
                missed.append(json_path)
            else:
                with open(json_path) as json_file:
                    json_config = yaml.safe_load(
                        self.rapi.task.render_template(
                            task_template=json_file.read()))
                with open(yaml_path) as yaml_file:
                    yaml_config = yaml.safe_load(
                        self.rapi.task.render_template(
                            task_template=yaml_file.read()))
                if json_config != yaml_config:
                    not_equal.append("'%s' and '%s'" % (yaml_path, json_path))

        error = ""
        if not_equal:
            error += ("Sample task configs are not equal:\n\t%s\n"
                      % "\n\t".join(not_equal))
        if missed:
            self.fail("Sample task configs are missing:\n\t%s\n"
                      % "\n\t".join(missed))

        if error:
            self.fail(error)

    def test_no_underscores_in_filename(self):
        bad_filenames = []
        for dirname, dirnames, filenames in os.walk(self.samples_path):
            for filename in filenames:
                if "_" in filename and (
                        filename.endswith(".yaml")
                        or filename.endswith(".json")):
                    full_path = os.path.join(dirname, filename)
                    bad_filenames.append(full_path)

        self.assertEqual([], bad_filenames,
                         "Following sample task filenames contain "
                         "underscores (_) but must use dashes (-) instead: "
                         "{}".format(bad_filenames))

    def test_context_samples_found(self):
        all_plugins = context.Context.get_all()
        context_samples_path = os.path.join(self.samples_path, "contexts")
        for p in all_plugins:
            # except contexts which belongs to tests module
            if not inspect.getfile(p).startswith(
               os.path.dirname(rally.__file__)):
                continue
            file_name = p.get_name().replace("_", "-")
            file_path = os.path.join(context_samples_path, file_name)
            if not os.path.exists("%s.json" % file_path):
                self.fail(("There is no json sample file of %s,"
                           "plugin location: %s" %
                           (p.get_name(), p.__module__)))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import fnmatch
import io
import os
import re

import testtools


class TestFormat(testtools.TestCase):
    def _check_lines_wrapping(self, doc_file, raw):
        code_block = False
        text_inside_simple_tables = False
        lines = raw.split("\n")
        for i, line in enumerate(lines):
            if code_block:
                if not line or line.startswith(" "):
                    continue
                else:
                    code_block = False
            if "::" in line:
                code_block = True
            # simple style tables also can fit >=80 symbols
            # open simple style table
            if ("===" in line or "---" in line) and not lines[i - 1]:
                text_inside_simple_tables = True
            if "http://" in line or "https://" in line or ":ref:" in line:
                continue
            # Allow lines which do not contain any whitespace
            if re.match(r"\s*[^\s]+$", line):
                continue
            if not text_inside_simple_tables:
                self.assertTrue(
                    len(line) < 80,
                    msg="%s:%d: Line limited to a maximum of 79 characters." %
                    (doc_file, i + 1))
            # close simple style table
            if "===" in line and not lines[i + 1]:
                text_inside_simple_tables = False

    def _check_no_cr(self, doc_file, raw):
        matches = re.findall("\r", raw)
        self.assertEqual(
            len(matches), 0,
            "Found %s literal carriage returns in file %s" %
            (len(matches), doc_file))

    def _check_trailing_spaces(self, doc_file, raw):
        for i, line in enumerate(raw.split("\n")):
            trailing_spaces = re.findall(r"\s+$", line)
            self.assertEqual(
                len(trailing_spaces), 0,
                "Found trailing spaces on line %s of %s" % (i + 1, doc_file))

    def test_lines(self):

        files = []
        docs_dir = os.path.join(os.path.dirname(__file__), os.pardir,
                                os.pardir, os.pardir, "doc")
        for root, dirnames, filenames in os.walk(docs_dir):
            for filename in fnmatch.filter(filenames, "*.rst"):
                files.append(os.path.join(root, filename))

        for filename in files:
            with io.open(filename, encoding="utf-8") as f:
                data = f.read()

            self._check_lines_wrapping(filename, data)
            self._check_no_cr(filename, data)
            self._check_trailing_spaces(filename, data)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import importlib

from rally.common import cfg
from rally.common import logging
from rally.task import engine

CONF = cfg.CONF


def list_opts():

    merged_opts = {"DEFAULT": []}
    merged_opts["DEFAULT"].extend(logging.DEBUG_OPTS)
    merged_opts["DEFAULT"].extend(engine.TASK_ENGINE_OPTS)

    return merged_opts.items()


def update_opt_defaults():
    logging.oslogging.cfg.set_defaults(
        logging.oslogging._options.generic_log_opts,
        use_stderr=True
    )


_registered = False
_registered_paths = []


def register_options_from_path(path):
    global _registered_paths
    if path not in _registered_paths:
        if ":" not in path:
            return
        module_name, function_name = path.split(":", 1)
        try:
            module = importlib.import_module(module_name)
        except ImportError:
            return
        list_func = getattr(module, function_name, None)
        if list_func is None:
            return

        options = list_func()

        if isinstance(options, dict):
            options = options.items()

        register_opts(options)
        _registered_paths.append(path)


def register_opts(opts):
    for category, options in opts:
        group = cfg.OptGroup(name=category, title="%s options" % category)
        if category != "DEFAULT":
            CONF.register_group(group)
            CONF.register_opts(options, group=group)
        else:
            CONF.register_opts(options)


def register():
    global _registered

    if not _registered:
        register_opts(list_opts())

        _registered = True



# Copyright 2015: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy


class MetaMixin(object):
    """Safe way to store meta information related to class object.

    Allows to store information in class object instead of the instance.

    Information is stored in dict that is initialized only once during the
    load of module, it means that all subclasses of this class will point to
    the same dict object with the information.

    Allows to protect children from using parents meta.

    Sample that explains why it's important to use MetaMixin:

        >>> # Using direct fields
        >>>
        >>> class A(object):
        >>>     _meta = {}
        >>>
        >>> class B(A):
        >>>    pass
        >>>
        >>> B._meta["a"] = 10
        >>> assert A._meta["a"] == 10  # We changed meta of base class, which
                                       # is going to produce nasty bugs

        >>> # MetaMixin in action
        >>>
        >>> class A(MetaMixin):
        >>>    pass
        >>>
        >>> class B(A):
        >>>    pass
        >>>
        >>> A._meta_set("a", 10)   # Raises ReferenceError
        >>> A._meta_init()
        >>> A._meta_set("a", 10)   # Set meta field "a"
        >>>
        >>> B._meta_get("a")       # Raises ReferenceError
        >>> B._meta_init()
        >>> B._meta_set("a", 20)   # Set meta field "a"
        >>>
        >>> assert A._meta_get("a")  == 10
        >>> assert B._meta_get("a")  == 20
    """

    _default_meta = (None, {})

    @classmethod
    def _meta_init(cls):
        """Initialize meta for this class."""
        cls._meta = copy.deepcopy(cls._default_meta[1])

    @classmethod
    def _meta_clear(cls):
        cls._meta.clear()    # NOTE(boris-42): make sure that meta is deleted
        delattr(cls, "_meta")

    @classmethod
    def _meta_is_inited(cls, raise_exc=True):
        """Check if meta is initialized.

        It means that this class has own cls._meta object (not pointer
        to parent cls._meta)
        """
        if vars(cls).get("_meta") is None:
            if raise_exc:
                raise ReferenceError(
                    "Trying to use MetaMixin before initialization %s. "
                    "Call _meta_init() before using it" % cls)
            return False
        return True

    @classmethod
    def _meta_get(cls, key, default=None):
        """Get value corresponding to key in meta data."""
        cls._meta_is_inited()
        return cls._meta.get(key, default)

    @classmethod
    def _meta_set(cls, key, value):
        """Set value for key in meta."""
        cls._meta_is_inited()
        cls._meta[key] = value

    @classmethod
    def _meta_setdefault(cls, key, value):
        """Set default value for key in meta."""
        cls._meta_is_inited()
        cls._meta.setdefault(key, value)

    @classmethod
    def _default_meta_init(cls, inherit=True):
        """Initialize default meta.

        Default Meta is used to change the behavior of _meta_init() method
        Meta is initialized with the copy of default meta instead of {}

        :param inherit: initialize meta with copy of parent's default meta
        """
        if inherit:
            cls._default_meta = (cls, copy.deepcopy(cls._default_meta[1]))
        else:
            cls._default_meta = (cls, {})

    @classmethod
    def _default_meta_set(cls, key, value):
        if cls is not cls._default_meta[0]:
            raise ReferenceError(
                "Trying to update default meta from children class.")

        cls._default_meta[1][key] = value

    @classmethod
    def _default_meta_get(cls, key, default=None):
        return cls._default_meta[1].get(key, default)

    @classmethod
    def _default_meta_setdefault(cls, key, value):
        if cls is not cls._default_meta[0]:
            raise ReferenceError(
                "Trying to update default meta from children class.")

        cls._default_meta[1].setdefault(key, value)



# Copyright 2015: Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

import testtools

from rally import exceptions
from rally.task import functional
from tests.unit import test


class FunctionalMixinTestCase(test.TestCase):

    def test_asserts(self):
        class A(functional.FunctionalMixin):
            def __init__(self):
                super(A, self).__init__()

        a = A()
        a.assertEqual(1, 1)
        self.assertRaises(exceptions.RallyAssertionError,
                          a.assertEqual, "a", "b")

        a.assertNotEqual(1, 2)
        self.assertRaises(exceptions.RallyAssertionError,
                          a.assertNotEqual, "a", "a")

        a.assertTrue(True)
        self.assertRaises(exceptions.RallyAssertionError,
                          a.assertTrue, False)

        a.assertFalse(False)
        self.assertRaises(exceptions.RallyAssertionError,
                          a.assertFalse, True)

        a.assertIs("a", "a")
        self.assertRaises(exceptions.RallyAssertionError,
                          a.assertIs, "a", "b")

        a.assertIsNot("a", "b")
        self.assertRaises(exceptions.RallyAssertionError,
                          a.assertIsNot, "a", "a")

        a.assertIsNone(None)
        self.assertRaises(exceptions.RallyAssertionError,
                          a.assertIsNone, "a")

        a.assertIsNotNone("a")
        self.assertRaises(exceptions.RallyAssertionError,
                          a.assertIsNotNone, None)

        a.assertIn("1", ["1", "2", "3"])
        self.assertRaises(exceptions.RallyAssertionError,
                          a.assertIn, "4", ["1", "2", "3"])

        a.assertNotIn("4", ["1", "2", "3"])
        self.assertRaises(exceptions.RallyAssertionError,
                          a.assertNotIn, "1", ["1", "2", "3"])

        a.assertIsSubset(["1", "2"], ["1", "2", "3"])
        self.assertRaises(exceptions.RallyAssertionError,
                          a.assertIsSubset, ["1", "4"], ["1", "2", "3"])

        a.assertIsNotSubset(["1", "4"], ["1", "2", "3"])
        self.assertRaises(exceptions.RallyAssertionError,
                          a.assertIsNotSubset, ["1", "2"], ["1", "2", "3"])

        a.assertIsInstance("a", str)
        self.assertRaises(exceptions.RallyAssertionError,
                          a.assertIsInstance, "a", int)

        a.assertIsNotInstance("a", int)
        self.assertRaises(exceptions.RallyAssertionError,
                          a.assertIsNotInstance, "a", str)

        a.assertLessEqual(len(["1", "2"]), len(["3", "4", "5"]))
        a.assertLessEqual(len(["1", "2", "3"]), len(["3", "4", "5"]))
        self.assertRaises(exceptions.RallyAssertionError,
                          a.assertLessEqual,
                          len(["1", "2", "3"]), len(["3", "4"]))

        a.assertLess(len(["1", "2"]), len(["3", "4", "5"]))
        self.assertRaises(exceptions.RallyAssertionError,
                          a.assertLess,
                          len(["1", "2", "3"]), len(["3", "4", "5"]))
        self.assertRaises(exceptions.RallyAssertionError,
                          a.assertLess, len(["1", "2", "3"]), len(["3", "4"]))

        a.assertGreaterEqual(len(["1", "2", "9", "8"]), len(["3", "4", "5"]))
        a.assertGreaterEqual(len(["1", "2", "3"]), len(["3", "4", "5"]))
        self.assertRaises(exceptions.RallyAssertionError,
                          a.assertGreaterEqual,
                          len(["1", "2"]), len(["3", "4", "5"]))

        a.assertGreater(len(["1", "2", "9", "8"]), len(["3", "4", "5"]))
        self.assertRaises(exceptions.RallyAssertionError,
                          a.assertGreater,
                          len(["1", "2", "3"]), len(["3", "4", "5"]))
        self.assertRaises(exceptions.RallyAssertionError,
                          a.assertGreater,
                          len(["1", "2"]), len(["3", "4", "5"]))

    @testtools.skipIf(sys.version_info < (2, 7),
                      "assertRaises as context not supported")
    def test_assert_with_custom_message(self):
        class A(functional.FunctionalMixin):
            def __init__(self):
                super(A, self).__init__()

        a = A()
        custom_message = "A custom message"
        assert_message = "Assertion error: .+\\. " + custom_message

        a.assertEqual(1, 1, "It's equal")
        message = self._catch_exception_message(a.assertEqual,
                                                "a", "b", custom_message)
        self.assertRegex(message, assert_message)

        a.assertNotEqual(1, 2)
        message = self._catch_exception_message(a.assertNotEqual,
                                                "a", "a", custom_message)
        self.assertRegex(message, assert_message)

        a.assertTrue(True)
        message = self._catch_exception_message(a.assertTrue,
                                                False, custom_message)
        self.assertRegex(message, assert_message)

        a.assertFalse(False)
        message = self._catch_exception_message(a.assertFalse,
                                                True, custom_message)
        self.assertRegex(message, assert_message)

        a.assertIs("a", "a")
        message = self._catch_exception_message(a.assertIs,
                                                "a", 1, custom_message)
        self.assertRegex(message, assert_message)

        a.assertIsNot("a", "b")
        message = self._catch_exception_message(a.assertIsNot,
                                                "a", "a", custom_message)
        self.assertRegex(message, assert_message)

        a.assertIsNone(None)
        message = self._catch_exception_message(a.assertIsNone,
                                                "a", custom_message)
        self.assertRegex(message, assert_message)

        a.assertIsNotNone("a")
        message = self._catch_exception_message(a.assertIsNotNone,
                                                None, custom_message)
        self.assertRegex(message, assert_message)

        a.assertIn("1", ["1", "2", "3"])
        message = self._catch_exception_message(a.assertIn,
                                                "1", ["2", "3", "4"],
                                                custom_message)
        self.assertRegex(message, assert_message)

        a.assertNotIn("4", ["1", "2", "3"])
        message = self._catch_exception_message(a.assertNotIn,
                                                "1", ["1", "2", "3"],
                                                custom_message)
        self.assertRegex(message, assert_message)

        a.assertIsInstance("a", str)
        message = self._catch_exception_message(a.assertIsInstance,
                                                "a", int, custom_message)
        self.assertRegex(message, assert_message)

        a.assertIsNotInstance("a", int)
        message = self._catch_exception_message(a.assertIsNotInstance,
                                                "a", str, custom_message)
        self.assertRegex(message, assert_message)

        a.assertLessEqual(len(["1", "2"]), len(["3", "4", "5"]))
        a.assertLessEqual(len(["1", "2", "3"]), len(["3", "4", "5"]))
        self.assertRaises(exceptions.RallyAssertionError,
                          a.assertLessEqual,
                          len(["1", "2", "3"]), len(["3", "4"]),
                          custom_message)

        a.assertLess(len(["1", "2"]), len(["3", "4", "5"]))
        self.assertRaises(exceptions.RallyAssertionError,
                          a.assertLess,
                          len(["1", "2", "3"]), len(["3", "4", "5"]),
                          custom_message)
        self.assertRaises(exceptions.RallyAssertionError,
                          a.assertLess, len(["1", "2", "3"]), len(["3", "4"]),
                          custom_message)

        a.assertGreaterEqual(len(["1", "2", "9", "8"]), len(["3", "4", "5"]))
        a.assertGreaterEqual(len(["1", "2", "3"]), len(["3", "4", "5"]))
        self.assertRaises(exceptions.RallyAssertionError,
                          a.assertGreaterEqual,
                          len(["1", "2"]), len(["3", "4", "5"]),
                          custom_message)

        a.assertGreater(len(["1", "2", "9", "8"]), len(["3", "4", "5"]))
        self.assertRaises(exceptions.RallyAssertionError,
                          a.assertGreater,
                          len(["1", "2", "3"]), len(["3", "4", "5"]),
                          custom_message)
        self.assertRaises(exceptions.RallyAssertionError,
                          a.assertGreater,
                          len(["1", "2"]), len(["3", "4", "5"]),
                          custom_message)

    def _catch_exception_message(self, func, *args):
        try:
            func(*args)
        except Exception as e:
            return str(e)



# Copyright 2016: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from rally.common import db
from rally import consts
from rally import exceptions
from rally.verification import manager


class Verifier(object):
    """Represents a verifier object."""
    TIME_FORMAT = consts.TimeFormat.ISO8601

    def __init__(self, verifier):
        """Init a verifier object.

        :param verifier: Dict representation of a verifier in the database
        """
        self._db_entry = verifier

        self._deployment = None
        self._manager = None

    def __getattr__(self, attr):
        return self._db_entry[attr]

    def __getitem__(self, item):
        return self._db_entry[item]

    def __str__(self):
        return "'%s' (UUID=%s)" % (self.name, self.uuid)

    def to_dict(self, item=None):
        data = {}
        formatters = ["created_at", "updated_at"]
        fields = ["status", "system_wide", "uuid", "type", "platform",
                  "name", "source", "version", "extra_settings",
                  "id", "description"]
        for field in fields:
            data[field] = self._db_entry.get(field, "")
        for field in formatters:
            data[field] = self._db_entry.get(field, "").strftime(
                self.TIME_FORMAT)
        return data

    @classmethod
    def create(cls, name, vtype, platform, source, version, system_wide,
               extra_settings=None):
        db_entry = db.verifier_create(name=name, vtype=vtype,
                                      platform=platform, source=source,
                                      version=version, system_wide=system_wide,
                                      extra_settings=extra_settings)
        return cls(db_entry)

    @classmethod
    def get(cls, verifier_id):
        return cls(db.verifier_get(verifier_id))

    @classmethod
    def list(cls, status=None):
        return [cls(db_entry) for db_entry in db.verifier_list(status)]

    @staticmethod
    def delete(verifier_id):
        db.verifier_delete(verifier_id)

    def update_status(self, status):
        self.update_properties(status=status)

    def update_properties(self, **properties):
        self._db_entry = db.verifier_update(self.uuid, **properties)

    def set_env(self, env_id):
        from rally.common import objects
        self._deployment = objects.Deployment.get(env_id)

    @property
    def deployment(self):
        # TODO(andreykurilin): deprecate this property someday
        if self._deployment is None:
            raise exceptions.RallyException(
                "Verifier is not linked to any deployment. Please, call "
                "`set_env` method.")
        return self._deployment

    @property
    def env(self):
        return self.deployment.env_obj

    @property
    def manager(self):
        # lazy load manager to be able to use non-plugin related stuff without
        # loading plugins
        if not self._manager:
            self._manager = manager.VerifierManager.get(
                self.type, self.platform)(self)
        return self._manager



# Copyright 2013: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""High level ssh library.

Usage examples:

Execute command and get output:

    ssh = sshclient.SSH("root", "example.com", port=33)
    status, stdout, stderr = ssh.execute("ps ax")
    if status:
        raise Exception("Command failed with non-zero status.")
    print stdout.splitlines()

Execute command with huge output:

    class PseudoFile(object):
        def write(chunk):
            if "error" in chunk:
                email_admin(chunk)

    ssh = sshclient.SSH("root", "example.com")
    ssh.run("tail -f /var/log/syslog", stdout=PseudoFile(), timeout=False)

Execute local script on remote side:

    ssh = sshclient.SSH("user", "example.com")
    status, out, err = ssh.execute("/bin/sh -s arg1 arg2",
                                   stdin=open("~/myscript.sh", "r"))

Upload file:

    ssh = sshclient.SSH("user", "example.com")
    ssh.run("cat > ~/upload/file.gz", stdin=open("/store/file.gz", "rb"))

Eventlet:

    eventlet.monkey_patch(select=True, time=True)
    or
    eventlet.monkey_patch()
    or
    sshclient = eventlet.import_patched("opentstack.common.sshclient")

"""

import io
import os
import select
import shlex
import socket
import time

import paramiko

from rally.common import logging
from rally import exceptions

LOG = logging.getLogger(__name__)


class SSH(object):
    """Represent ssh connection."""

    def __init__(self, user, host, port=22, pkey=None,
                 key_filename=None, password=None):
        """Initialize SSH client.

        :param user: ssh username
        :param host: hostname or ip address of remote ssh server
        :param port: remote ssh port
        :param pkey: RSA or DSS private key string or file object
        :param key_filename: private key filename
        :param password: password
        """

        self.user = user
        self.host = host
        self.port = port
        self.pkey = self._get_pkey(pkey) if pkey else None
        self.password = password
        self.key_filename = key_filename
        self._client = False

    def _get_pkey(self, key):
        if isinstance(key, str):
            key = io.StringIO(key)
        errors = []
        key_pos = key.seek(0, 1)
        for key_class in (paramiko.rsakey.RSAKey, paramiko.dsskey.DSSKey):
            try:
                return key_class.from_private_key(key)
            except paramiko.SSHException as e:
                key.seek(key_pos)
                errors.append(e)
        raise exceptions.SSHError("Invalid pkey: %s" % (errors))

    def _get_client(self):
        if self._client:
            return self._client
        try:
            self._client = paramiko.SSHClient()
            self._client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            self._client.connect(self.host, username=self.user,
                                 port=self.port, pkey=self.pkey,
                                 key_filename=self.key_filename,
                                 password=self.password, timeout=1)
            return self._client
        except Exception as e:
            message = ("Exception %(exception_type)s was raised "
                       "during connect to %(user)s@%(host)s:%(port)s. "
                       "Exception value is: %(exception)r")
            self._client = False
            raise exceptions.SSHError(message % {"exception": e,
                                                 "user": self.user,
                                                 "host": self.host,
                                                 "port": self.port,
                                                 "exception_type": type(e)})

    def close(self):
        self._client.close()
        self._client = False

    def run(self, cmd, stdin=None, stdout=None, stderr=None,
            raise_on_error=True, timeout=3600):
        """Execute specified command on the server.

        :param cmd:             Command to be executed.
        :param stdin:           Open file or string to pass to stdin.
        :param stdout:          Open file to connect to stdout.
        :param stderr:          Open file to connect to stderr.
        :param raise_on_error:  If False then exit code will be return. If True
                                then exception will be raised if non-zero code.
        :param timeout:         Timeout in seconds for command execution.
                                Default 1 hour. No timeout if set to 0.
        """

        client = self._get_client()

        if isinstance(stdin, str):
            stdin = io.StringIO(stdin)

        return self._run(client, cmd, stdin=stdin, stdout=stdout,
                         stderr=stderr, raise_on_error=raise_on_error,
                         timeout=timeout)

    def _run(self, client, cmd, stdin=None, stdout=None, stderr=None,
             raise_on_error=True, timeout=3600):

        if isinstance(cmd, (list, tuple)):
            cmd = " ".join(shlex.quote(str(p)) for p in cmd)

        transport = client.get_transport()
        session = transport.open_session()
        session.exec_command(cmd)
        start_time = time.time()

        data_to_send = ""
        stderr_data = None

        # If we have data to be sent to stdin then `select' should also
        # check for stdin availability.
        if stdin and not stdin.closed:
            writes = [session]
        else:
            writes = []

        data = None
        while True:
            # Block until data can be read/write.
            r, w, e = select.select([session], writes, [session], 1)

            if session.recv_ready():
                data = session.recv(4096)
                LOG.debug("stdout: %r" % data)
                if stdout is not None:
                    stdout.write(data.decode("utf8"))
                continue

            if session.recv_stderr_ready():
                stderr_data = session.recv_stderr(4096)
                LOG.debug("stderr: %r" % stderr_data)
                if stderr is not None:
                    stderr.write(stderr_data.decode("utf8"))
                continue

            if session.send_ready():
                if stdin is not None and not stdin.closed:
                    if not data_to_send:
                        data_to_send = stdin.read(4096)
                        if not data_to_send:
                            stdin.close()
                            session.shutdown_write()
                            writes = []
                            continue
                    sent_bytes = session.send(data_to_send)
                    LOG.debug("sent: %s" % data_to_send[:sent_bytes])
                    data_to_send = data_to_send[sent_bytes:]

            if session.exit_status_ready():
                break

            if timeout and (time.time() - timeout) > start_time:
                args = {"cmd": cmd, "host": self.host}
                raise exceptions.SSHTimeout("Timeout executing command "
                                            "'%(cmd)s' on host %(host)s"
                                            % args)
            if e:
                raise exceptions.SSHError("Socket error.")

        exit_status = session.recv_exit_status()
        if 0 != exit_status and raise_on_error:
            fmt = "Command '%(cmd)s' failed with exit_status %(status)d."
            details = fmt % {"cmd": cmd, "status": exit_status}
            if stderr_data:
                details += " Last stderr data: '%s'." % stderr_data
            raise exceptions.SSHError(details)
        return exit_status, data

    def execute(self, cmd, stdin=None, timeout=3600):
        """Execute the specified command on the server.

        :param cmd:     Command to be executed, can be a list.
        :param stdin:   Open file to be sent on process stdin.
        :param timeout: Timeout for execution of the command.

        :returns: tuple (exit_status, stdout, stderr)
        """
        stdout = io.StringIO()
        stderr = io.StringIO()

        exit_status, data = self.run(cmd, stderr=stderr, stdout=stdout,
                                     stdin=stdin, timeout=timeout,
                                     raise_on_error=False)
        stdout.seek(0)
        stderr.seek(0)
        return (exit_status, stdout.read(), stderr.read())

    def wait(self, timeout=120, interval=1):
        """Wait for the host will be available via ssh."""
        start_time = time.time()
        while True:
            try:
                return self.execute("uname")
            except (socket.error, exceptions.SSHError) as e:
                LOG.debug("Ssh is still unavailable: %r" % e)
                time.sleep(interval)
            if time.time() > (start_time + timeout):
                raise exceptions.SSHTimeout("Timeout waiting for '%s'" %
                                            self.host)

    def _put_file_sftp(self, localpath, remotepath, mode=None):
        client = self._get_client()

        with client.open_sftp() as sftp:
            sftp.put(localpath, remotepath)
            if mode is None:
                mode = 0o777 & os.stat(localpath).st_mode
            sftp.chmod(remotepath, mode)

    def _put_file_shell(self, localpath, remotepath, mode=None):
        cmd = ["cat > %s" % remotepath]
        if mode is not None:
            cmd.append("chmod 0%o %s" % (mode, remotepath))

        with open(localpath, "rb") as localfile:
            cmd = "; ".join(cmd)
            self.run(cmd, stdin=localfile)

    def put_file(self, localpath, remotepath, mode=None):
        """Copy specified local file to the server.

        :param localpath:   Local filename.
        :param remotepath:  Remote filename.
        :param mode:        Permissions to set after upload
        """
        import socket
        try:
            self._put_file_sftp(localpath, remotepath, mode=mode)
        except (paramiko.SSHException, socket.error):
            self._put_file_shell(localpath, remotepath, mode=mode)



# Copyright (C) 2014 Yahoo! Inc. All Rights Reserved.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from rally.common import utils as rutils
from rally import consts
from rally.task import runner


@runner.configure(name="serial")
class SerialScenarioRunner(runner.ScenarioRunner):
    """Scenario runner that executes scenarios serially.

    Unlike scenario runners that execute in parallel, the serial scenario
    runner executes scenarios one-by-one in the same python interpreter process
    as Rally. This allows you to execute scenario without introducing
    any concurrent operations as well as interactively debug the scenario
    from the same command that you use to start Rally.
    """

    # NOTE(mmorais): additionalProperties is set True to allow switching
    # between parallel and serial runners by modifying only *type* property
    CONFIG_SCHEMA = {
        "type": "object",
        "$schema": consts.JSON_SCHEMA,
        "properties": {
            "times": {
                "type": "integer",
                "minimum": 1
            }
        },
        "additionalProperties": True
    }

    def _run_scenario(self, cls, method_name, context, args):
        """Runs the specified scenario with given arguments.

        The scenario iterations are executed one-by-one in the same python
        interpreter process as Rally. This allows you to execute
        scenario without introducing any concurrent operations as well as
        interactively debug the scenario from the same command that you use
        to start Rally.

        :param cls: The Scenario class where the scenario is implemented
        :param method_name: Name of the method that implements the scenario
        :param context: context that contains users, admin & other
                        information, that was created before scenario
                        execution starts.
        :param args: Arguments to call the scenario method with

        :returns: List of results fore each single scenario iteration,
                  where each result is a dictionary
        """
        times = self.config.get("times", 1)

        event_queue = rutils.DequeAsQueue(self.event_queue)

        for i in range(times):
            if self.aborted.is_set():
                break
            result = runner._run_scenario_once(
                cls, method_name, runner._get_scenario_context(i, context),
                args, event_queue)
            self._send_result(result)

        self._flush_results()



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import jsonschema

from rally.env import env_mgr
from rally.env import platform

from tests.unit import test


class PlatformBaseTestCase(test.TestCase):

    def _check_schema(self, schema, obj):
        jsonschema.validate(obj, schema)

    def _check_health_schema(self, obj):
        self._check_schema(env_mgr.EnvManager._HEALTH_FORMAT, obj)

    def _check_cleanup_schema(self, obj):
        self._check_schema(env_mgr.EnvManager._CLEANUP_FORMAT, obj)

    def _check_info_schema(self, obj):
        self._check_schema(env_mgr.EnvManager._INFO_FORMAT, obj)


class PlatformTestCase(test.TestCase):

    def test_plugin_configure_and_methods(self):

        @platform.configure(name="existing", platform="foo")
        class FooPlugin(platform.Platform):
            pass

        self.addCleanup(FooPlugin.unregister)

        f = FooPlugin("spec", "uuid", "plugin_data", "platform_data", "status")
        self.assertEqual(f.uuid, "uuid")
        self.assertEqual(f.spec, "spec")
        self.assertEqual(f.plugin_data, "plugin_data")
        self.assertEqual(f.platform_data, "platform_data")
        self.assertEqual(f.status, "status")

        self.assertRaises(NotImplementedError, f.create)
        self.assertRaises(NotImplementedError, f.destroy)
        self.assertRaises(NotImplementedError, f.update, "new_spec")
        self.assertRaises(NotImplementedError, f.cleanup)
        self.assertRaises(NotImplementedError,
                          f.cleanup, task_uuid="task_uuid")
        self.assertRaises(NotImplementedError, f.check_health)
        self.assertRaises(NotImplementedError, f.info)



# Copyright 2017: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import mock

from rally.cli.commands import db
from tests.unit import fakes
from tests.unit import test


class DBCommandsTestCase(test.TestCase):

    def setUp(self):
        super(DBCommandsTestCase, self).setUp()
        self.db_commands = db.DBCommands()
        self.fake_api = fakes.FakeAPI()

    @mock.patch("rally.cli.commands.db.envutils")
    @mock.patch("rally.cli.commands.db.db.schema")
    def test_recreate(self, mock_db_schema, mock_envutils):
        self.db_commands.recreate(self.fake_api)
        db_calls = [mock.call.schema_cleanup(),
                    mock.call.schema_create()]
        self.assertEqual(db_calls, mock_db_schema.mock_calls)
        envutils_calls = [mock.call.clear_env()]
        self.assertEqual(envutils_calls, mock_envutils.mock_calls)

    @mock.patch("rally.cli.commands.db.db.schema")
    def test_create(self, mock_db_schema):
        self.db_commands.create(self.fake_api)
        calls = [mock.call.schema_create()]
        self.assertEqual(calls, mock_db_schema.mock_calls)

    @mock.patch("rally.cli.commands.db.db.schema")
    def test_ensure_create(self, mock_db_schema):
        mock_db_schema.schema_revision.return_value = None
        self.db_commands.ensure(self.fake_api)
        calls = [mock.call.schema_revision(),
                 mock.call.schema_create()]
        self.assertEqual(calls, mock_db_schema.mock_calls)

    @mock.patch("rally.cli.commands.db.db.schema")
    def test_ensure_exists(self, mock_db_schema):
        mock_db_schema.schema_revision.return_value = "revision"
        self.db_commands.ensure(self.fake_api)
        calls = [mock.call.schema_revision()]
        self.assertEqual(calls, mock_db_schema.mock_calls)

    @mock.patch("rally.cli.commands.db.db.schema")
    def test_upgrade(self, mock_db_schema):
        self.db_commands.upgrade(self.fake_api)
        calls = [mock.call.schema_upgrade()]
        mock_db_schema.assert_has_calls(calls)

    @mock.patch("rally.cli.commands.db.db.schema")
    def test_revision(self, mock_db_schema):
        self.db_commands.revision(self.fake_api)
        calls = [mock.call.schema_revision()]
        mock_db_schema.assert_has_calls(calls)

    @mock.patch("rally.cli.commands.db.print")
    @mock.patch("rally.cli.commands.db.cfg.CONF.database")
    def test_show(self, mock_conf_database, mock_print):
        mock_conf_database.connection = "http://aaa:bbb@testing.com:888"
        self.db_commands.show(self.fake_api)
        mock_print.assert_called_once_with("http://**:**@testing.com:888")
        mock_print.reset_mock()
        self.db_commands.show(self.fake_api, show_creds=True)
        mock_print.assert_called_once_with("http://aaa:bbb@testing.com:888")



# Copyright 2013: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from rally.common.plugin import discover


_exception_map = None


class RallyException(Exception):
    """Base Rally Exception

    To correctly use this class, inherit from it and define
    a "msg_fmt" property. That msg_fmt will get printf'd
    with the keyword arguments provided to the constructor.

    """
    msg_fmt = "%(message)s"
    error_code = 100

    def __init__(self, message=None, **kwargs):
        self.kwargs = kwargs

        if "%(message)s" in self.msg_fmt:
            kwargs.update({"message": message})

        super(RallyException, self).__init__(self.msg_fmt % kwargs)

    def format_message(self):
        return str(self)


def find_exception(response):
    """Discover a proper exception class based on response object."""
    global _exception_map
    if _exception_map is None:
        _exception_map = dict(
            (e.error_code, e) for e in discover.itersubclasses(RallyException))
    exc_class = _exception_map.get(response.status_code, RallyException)

    error_data = response.json()["error"]
    if error_data["args"]:
        return exc_class(error_data["args"])
    return exc_class(error_data["msg"])


def make_exception(exc):
    """Check a class of exception and convert it to rally-like if needed."""
    if isinstance(exc, RallyException):
        return exc
    return RallyException(str(exc))


class ValidationError(RallyException):
    error_code = 110
    msg_fmt = "Validation error: %(message)s"


class InvalidArgumentsException(RallyException):
    error_code = 111
    msg_fmt = "Invalid arguments: '%(message)s'"


class InvalidConfigException(RallyException):
    error_code = 112
    msg_fmt = "This config has invalid schema: `%(message)s`"


class InvalidTaskException(InvalidConfigException):
    error_code = 113
    msg_fmt = "Task config is invalid: `%(message)s`"


class InvalidTaskConfig(InvalidTaskException):
    error_code = 114
    msg_fmt = ("Input task is invalid!\n\n"
               "Subtask %(name)s[%(pos)s] has wrong configuration"
               "\nSubtask configuration:\n%(config)s\n"
               "\nReason(s):\n %(reason)s")


class InvalidScenarioArgument(RallyException):
    error_code = 115
    msg_fmt = "Invalid scenario argument: '%(message)s'"


class InvalidConnectionString(RallyException):
    error_code = 116
    msg_fmt = "Invalid connection string: %(message)s."


class DBException(RallyException):
    error_code = 200
    msg_fmt = "DB Exception: '%(message)s'"


class DBMigrationError(DBException):
    error_code = 201
    msg_fmt = "DB Migration Error: '%(message)s'"


class DBConflict(RallyException):
    error_code = 202
    msg_fmt = "DB Conflict. %(message)s"


class DBRecordNotFound(RallyException):
    error_code = 203
    msg_fmt = "Record for %(criteria)s not found in table %(table)s"


class DBRecordExists(DBException):
    error_code = 204
    msg_fmt = "Record with %(field)s = %(value)s already exists in %(table)s"


class DowngradeNotSupported(RallyException):
    error_code = 205
    msg_fmt = "Database schema downgrade is not supported."


class NotFoundException(RallyException):
    error_code = 210
    msg_fmt = "The resource can not be found: %(message)s"


class PluginNotFound(NotFoundException):
    error_code = 211
    msg_fmt = "There is no%(base)s plugin `%(name)s` in %(platform)s " \
              "platform."


class PluginWithSuchNameExists(RallyException):
    error_code = 212
    msg_fmt = (
        "Plugin with such name: %(name)s already exists in %(platform)s "
        "platform. It's module allocates at %(existing_path)s. You are trying "
        "to add plugin whose module allocates at %(new_path)s.")


class MultiplePluginsFound(RallyException):
    error_code = 213

    msg_fmt = ("Multiple plugins found: %(plugins)s for name %(name)s. "
               "Use full name with platform to fix issue.")


class GetResourceFailure(RallyException):
    error_code = 214
    msg_fmt = "Failed to get the resource %(resource)s: %(err)s"


class GetResourceNotFound(GetResourceFailure):
    error_code = 215
    msg_fmt = "Resource %(resource)s is not found."


class GetResourceErrorStatus(GetResourceFailure):
    error_code = 216
    msg_fmt = "Resource %(resource)s has %(status)s status.\n Fault: %(fault)s"


class AuthenticationFailed(InvalidArgumentsException):
    error_code = 220
    msg_fmt = ("Failed to authenticate to %(url)s for user '%(username)s'"
               " in project '%(project)s': %(etype)s: %(error)s")


class InvalidAdminException(InvalidArgumentsException):
    error_code = 221
    msg_fmt = "user '%(username)s' doesn't have 'admin' role"


class DeploymentNotFinishedStatus(RallyException):
    error_code = 222
    msg_fmt = "Deployment '%(name)s' (UUID=%(uuid)s) is '%(status)s'."


class ContextSetupFailure(RallyException):
    error_code = 223
    msg_fmt = "Unable to setup context '%(ctx_name)s': '%(msg)s'"


class ManagerException(RallyException):
    error_code = 230
    msg_fmt = "Internal error: %(message)s"


class ManagerInvalidSpec(ManagerException):
    error_code = 231
    msg_fmt = "%(mgr)s manager got invalid spec: \n%(errors)s"


class ManagerInvalidState(ManagerException):
    error_code = 232
    msg_fmt = ("%(mgr)s manager in invalid state "
               "expected `%(expected)s' actual `%(actual)s' ")


class TimeoutException(RallyException):
    error_code = 240
    msg_fmt = ("Rally tired waiting %(timeout).2f seconds for "
               "%(resource_type)s %(resource_name)s:%(resource_id)s to "
               "become %(desired_status)s current status %(resource_status)s")


class ThreadTimeoutException(RallyException):
    error_code = 241
    msg_fmt = "Iteration interrupted due to timeout."


class SSHTimeout(RallyException):
    error_code = 242
    pass


class SSHError(RallyException):
    error_code = 243
    pass


class RallyAssertionError(RallyException):
    msg_fmt = "Assertion error: %(message)s"


class ScriptError(RallyException):
    msg_fmt = "Script execution failed: %(message)s"



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""fill_missed_workload_info

Absorbed by fab4f4f31f8a_fill_missed_workload_info

Revision ID: c517b0011857
Revises: 35fe16d4ab1c
Create Date: 2017-06-22 18:46:09.281312

"""

from rally import exceptions

# revision identifiers, used by Alembic.
revision = "c517b0011857"
down_revision = "35fe16d4ab1c"
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    raise exceptions.DowngradeNotSupported()
